from deprecated.sphinx import deprecated
from spacepackets.ecss import PusTelecommand, PusService
from tmtccmd.pus.s11_tc_sched import Subservice


def __generic_param_less_tc_sched_cmd(
    subservice: int, apid: int = 0, seq_count: int = 0
) -> PusTelecommand:
    return PusTelecommand(
        service=PusService.S11_TC_SCHED,
        subservice=subservice,
        apid=apid,
        seq_count=seq_count,
    )


@deprecated(
    version="4.0.0a2",
    reason="use create... API instead",
)
def generate_enable_tc_sched_cmd(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return create_enable_tc_sched_cmd(apid, seq_count)


def create_enable_tc_sched_cmd(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return __generic_param_less_tc_sched_cmd(
        subservice=Subservice.TC_ENABLE, apid=apid, seq_count=seq_count
    )


@deprecated(
    version="4.0.0a2",
    reason="use create... API instead",
)
def generate_disable_tc_sched_cmd(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return create_disable_tc_sched_cmd(apid, seq_count)


def create_disable_tc_sched_cmd(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return __generic_param_less_tc_sched_cmd(
        subservice=Subservice.TC_DISABLE, apid=apid, seq_count=seq_count
    )


@deprecated(
    version="4.0.0a2",
    reason="use create... API instead",
)
def generate_reset_tc_sched_cmd(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return create_reset_tc_sched_cmd(apid, seq_count)


def create_reset_tc_sched_cmd(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return __generic_param_less_tc_sched_cmd(
        subservice=Subservice.TC_RESET, apid=apid, seq_count=seq_count
    )


@deprecated(
    version="4.0.0a2",
    reason="use create... API instead",
)
def generate_time_tagged_cmd(release_time: bytes, tc_to_insert: PusTelecommand):
    return create_time_tagged_cmd(release_time, tc_to_insert)


def create_time_tagged_cmd(
    release_time: bytes, tc_to_insert: PusTelecommand, apid: int = 0, seq_count: int = 0
):
    return PusTelecommand(
        service=PusService.S11_TC_SCHED,
        subservice=Subservice.TC_INSERT,
        app_data=pack_time_tagged_tc_app_data(release_time, tc_to_insert),
        apid=apid,
        seq_count=seq_count,
    )


def pack_time_tagged_tc_app_data(release_time: bytes, tc_to_insert: PusTelecommand) -> bytes:
    """This function packs another TC into an insert activity TC[11,4]
    :param release_time: Absolute time when TC shall be released/run
    :param tc_to_insert: The TC which shall be inserted
    """
    app_data = bytearray()
    # pack the release time
    app_data.extend(release_time)
    # followed by the tc
    app_data.extend(tc_to_insert.pack())
    return app_data



from collections import deque
from datetime import timedelta
from typing import cast
from unittest import TestCase
from unittest.mock import MagicMock

from spacepackets.ecss import PusTelecommand
from spacepackets.seqcount import ProvidesSeqCount


# Required for eval calls
# noinspection PyUnresolvedReferences
from tmtccmd.tmtc import (  # noqa: F401
    LogQueueEntry,
    QueueEntryHelper,
    RawTcEntry,
    WaitEntry,
)
from tmtccmd.tmtc.procedure import TreeCommandingProcedure
from tmtccmd.tmtc.queue import DefaultPusQueueHelper, QueueWrapper


class TestTcQueue(TestCase):
    def setUp(self) -> None:
        self.apid = 0x11
        self.queue_wrapper = QueueWrapper(info=TreeCommandingProcedure.empty(), queue=deque())
        self.assertEqual(self.queue_wrapper.queue, deque())
        self.queue_helper = DefaultPusQueueHelper(
            self.queue_wrapper,
            tc_sched_timestamp_len=4,
            seq_cnt_provider=None,
            default_pus_apid=None,
            pus_verificator=None,
        )
        self.pus_cmd = PusTelecommand(apid=self.apid, service=17, subservice=1)

    def test_wait_entry(self):
        self.queue_helper.add_wait(timedelta(seconds=2))
        self.assertEqual(len(self.queue_wrapper.queue), 1)
        wait_entry = cast(WaitEntry, self.queue_wrapper.queue.pop())
        self.assertTrue(wait_entry)
        self.assertFalse(wait_entry.is_tc())
        self.assertEqual(wait_entry.wait_time.total_seconds(), 2.0)
        self.assertEqual(len(self.queue_wrapper.queue), 0)

    def test_apid_stamper(self):
        self.queue_helper.pus_apid = 0x22
        self.queue_helper.add_pus_tc(self.pus_cmd)
        cast_wrapper = QueueEntryHelper(self.queue_wrapper.queue.popleft())
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc, self.pus_cmd)
        self.assertTrue(pus_entry)
        self.assertEqual(pus_entry.pus_tc.apid, 0x22)

    def test_seq_count_stamper(self):
        self.queue_helper.seq_cnt_provider = MagicMock(spec=ProvidesSeqCount)
        self.queue_helper.seq_cnt_provider.get_and_increment.return_value = 5
        self.queue_helper.add_pus_tc(self.pus_cmd)
        cast_wrapper = QueueEntryHelper(self.queue_wrapper.queue.popleft())
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc.seq_count, 5)

    def test_faulty_cast(self):
        self.queue_helper.add_pus_tc(self.pus_cmd)
        cast_wrapper = QueueEntryHelper(self.queue_wrapper.queue.popleft())
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc, self.pus_cmd)
        self.assertTrue(pus_entry)
        with self.assertRaises(TypeError):
            cast_wrapper.to_wait_entry()

    def test_multi_entry(self):
        pus_cmd = PusTelecommand(apid=self.apid, service=17, subservice=1)
        self.queue_helper.add_pus_tc(pus_cmd)
        self.assertEqual(len(self.queue_wrapper.queue), 1)
        self.queue_helper.add_log_cmd("Test String")
        self.queue_helper.add_raw_tc(bytes([0, 1, 2]))
        self.queue_helper.add_ccsds_tc(pus_cmd.to_space_packet())
        self.queue_helper.add_packet_delay(timedelta(seconds=3.0))
        self.assertEqual(len(self.queue_wrapper.queue), 5)
        pus_entry = self.queue_wrapper.queue.popleft()
        cast_wrapper = QueueEntryHelper(pus_entry)
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc, pus_cmd)
        self.assertTrue(pus_entry)
        log_entry = self.queue_wrapper.queue.popleft()
        self.assertFalse(log_entry.is_tc())
        cast_wrapper.entry = log_entry
        log_entry = cast_wrapper.to_log_entry()
        self.assertTrue(log_entry)
        with self.assertRaises(TypeError):
            cast_wrapper.to_raw_tc_entry()
        self.assertEqual(log_entry.log_str, "Test String")
        test_obj = eval(f"{log_entry!r}")
        self.assertEqual(test_obj.log_str, log_entry.log_str)

        raw_entry = self.queue_wrapper.queue.popleft()
        self.assertTrue(raw_entry.is_tc())
        cast_wrapper.entry = raw_entry
        raw_entry = cast_wrapper.to_raw_tc_entry()
        self.assertTrue(raw_entry)
        self.assertEqual(raw_entry.tc, bytes([0, 1, 2]))
        test_obj = eval(f"{raw_entry!r}")
        self.assertEqual(raw_entry.tc, test_obj.tc)

        space_packet_entry = self.queue_wrapper.queue.popleft()
        self.assertTrue(space_packet_entry.is_tc())
        cast_wrapper.entry = space_packet_entry
        space_packet_entry = cast_wrapper.to_space_packet_entry()
        self.assertTrue(space_packet_entry)
        self.assertTrue(space_packet_entry.space_packet, pus_cmd.to_space_packet())

        packet_delay = self.queue_wrapper.queue.pop()
        self.assertFalse(packet_delay.is_tc())
        cast_wrapper.entry = packet_delay
        packet_delay = cast_wrapper.to_packet_delay_entry()
        self.assertEqual(packet_delay.delay_time.total_seconds(), 3.0)



import logging
import os
from pathlib import Path
from unittest import TestCase

from tmtccmd.logging import add_colorlog_console_logger
from spacepackets.ecss import PusTelecommand
from spacepackets.ecss.pus_1_verification import (
    create_acceptance_success_tm,
    create_start_success_tm,
    create_step_success_tm,
    StepId,
    create_completion_success_tm,
    create_acceptance_failure_tm,
    FailureNotice,
    ErrorCode,
    create_start_failure_tm,
)
from spacepackets.ccsds.time import CdsShortTimestamp
from spacepackets.ecss.pus_verificator import PusVerificator
from tmtccmd.logging.pus import RegularTmtcLogWrapper
from tmtccmd.pus import VerificationWrapper


class TestPusVerifLog(TestCase):
    def setUp(self) -> None:
        self.apid = 0x02
        self.log_file_name = RegularTmtcLogWrapper.get_current_tmtc_file_name()
        self.logger = logging.getLogger(__name__)
        add_colorlog_console_logger(self.logger)

    def test_console_log_success(self):
        wrapper = VerificationWrapper(PusVerificator(), self.logger, None)
        self._test_success(wrapper)

    def test_console_log_success_without_colors(self):
        wrapper = VerificationWrapper(PusVerificator(), self.logger, None)
        wrapper.with_colors = False
        self._test_success(wrapper)

    def _test_success(self, wrapper: VerificationWrapper):
        verificator = wrapper.verificator
        tc = PusTelecommand(apid=self.apid, service=17, subservice=1, seq_count=0)
        verificator.add_tc(tc)
        srv_1_tm = create_acceptance_success_tm(
            apid=self.apid, pus_tc=tc, timestamp=CdsShortTimestamp.empty().pack()
        )
        empty_stamp = CdsShortTimestamp.empty().pack()

        def generic_checks():
            self.assertTrue("acc" in cm.output[0])
            self.assertTrue("fin" in cm.output[0])
            self.assertTrue("sta" in cm.output[0])
            self.assertTrue("ste" in cm.output[0])

        res = verificator.add_tm(srv_1_tm)
        with self.assertLogs(self.logger) as cm:
            wrapper.log_to_console(srv_1_tm, res)
            self.assertTrue(
                f"{'Acceptance success of TC'.ljust(25)} | "
                f"Request ID {srv_1_tm.tc_req_id.as_u32():#08x}" in cm.output[0]
            )
            generic_checks()
        srv_1_tm = create_start_success_tm(apid=self.apid, pus_tc=tc, timestamp=empty_stamp)
        res = verificator.add_tm(srv_1_tm)
        with self.assertLogs(self.logger) as cm:
            wrapper.log_to_console(srv_1_tm, res)
            self.assertTrue(
                f"{'Start success of TC'.ljust(25)} | "
                f"Request ID {srv_1_tm.tc_req_id.as_u32():#08x}" in cm.output[0]
            )
            generic_checks()
        srv_1_tm = create_step_success_tm(
            apid=self.apid,
            pus_tc=tc,
            step_id=StepId.with_byte_size(1, 1),
            timestamp=empty_stamp,
        )
        res = verificator.add_tm(srv_1_tm)
        with self.assertLogs(self.logger) as cm:
            wrapper.log_to_console(srv_1_tm, res)
            self.assertTrue(
                f"{'Step success of TC'.ljust(25)} | "
                f"Request ID {srv_1_tm.tc_req_id.as_u32():#08x}" in cm.output[0]
            )
            generic_checks()
        srv_1_tm = create_completion_success_tm(apid=self.apid, pus_tc=tc, timestamp=empty_stamp)
        res = verificator.add_tm(srv_1_tm)
        with self.assertLogs(self.logger) as cm:
            wrapper.log_to_console(srv_1_tm, res)
            self.assertTrue(
                f"{'Completion success of TC'.ljust(25)} | "
                f"Request ID {srv_1_tm.tc_req_id.as_u32():#08x}" in cm.output[0]
            )
            generic_checks()

    def test_console_log_acc_failure(self):
        wrapper = VerificationWrapper(PusVerificator(), self.logger, None)
        self._test_acc_failure(wrapper)

    def test_console_log_acc_failure_without_colors(self):
        wrapper = VerificationWrapper(PusVerificator(), self.logger, None)
        wrapper.with_colors = False
        self._test_acc_failure(wrapper)

    def _test_acc_failure(self, wrapper: VerificationWrapper):
        verificator = wrapper.verificator
        tc = PusTelecommand(apid=self.apid, service=17, subservice=1, seq_count=1)
        verificator.add_tc(tc)
        srv_1_tm = create_acceptance_failure_tm(
            apid=self.apid,
            pus_tc=tc,
            failure_notice=FailureNotice(code=ErrorCode(pfc=8, val=1), data=bytes()),
            timestamp=CdsShortTimestamp.empty().pack(),
        )
        res = verificator.add_tm(srv_1_tm)
        # TODO: Use self.assertLogs here instead
        wrapper.log_to_console(srv_1_tm, res)

    def test_console_log_start_failure(self):
        wrapper = VerificationWrapper(PusVerificator(), self.logger, None)
        verificator = wrapper.verificator
        tc = PusTelecommand(apid=self.apid, service=17, subservice=1, seq_count=2)
        verificator.add_tc(tc)
        srv_1_tm = create_acceptance_failure_tm(
            apid=self.apid,
            pus_tc=tc,
            failure_notice=FailureNotice(code=ErrorCode(pfc=8, val=1), data=bytes()),
            timestamp=CdsShortTimestamp.empty().pack(),
        )
        res = verificator.add_tm(srv_1_tm)
        # TODO: Use self.assertLogs here instead
        wrapper.log_to_console(srv_1_tm, res)
        srv_1_tm = create_start_failure_tm(
            apid=self.apid,
            pus_tc=tc,
            failure_notice=FailureNotice(code=ErrorCode(pfc=8, val=1), data=bytes()),
            timestamp=CdsShortTimestamp.empty().pack(),
        )
        res = verificator.add_tm(srv_1_tm)
        # TODO: Use self.assertLogs here instead
        wrapper.log_to_console(srv_1_tm, res)

    def test_file_logger(self):
        tmtc_logger = RegularTmtcLogWrapper(file_name=self.log_file_name)
        wrapper = VerificationWrapper(PusVerificator(), None, tmtc_logger.logger)
        verificator = wrapper.verificator
        tc = PusTelecommand(apid=self.apid, service=17, subservice=1, seq_count=0)
        verificator.add_tc(tc)
        srv_1_tm = create_acceptance_success_tm(
            apid=self.apid, pus_tc=tc, timestamp=CdsShortTimestamp.empty().pack()
        )
        res = verificator.add_tm(srv_1_tm)
        wrapper.log_to_file(srv_1_tm, res)
        srv_1_tm = create_start_success_tm(
            apid=self.apid, pus_tc=tc, timestamp=CdsShortTimestamp.empty().pack()
        )
        res = verificator.add_tm(srv_1_tm)
        wrapper.log_to_file(srv_1_tm, res)
        srv_1_tm = create_step_success_tm(
            apid=self.apid,
            pus_tc=tc,
            step_id=StepId.with_byte_size(1, 1),
            timestamp=CdsShortTimestamp.empty().pack(),
        )
        res = verificator.add_tm(srv_1_tm)
        wrapper.log_to_file(srv_1_tm, res)
        srv_1_tm = create_completion_success_tm(
            apid=self.apid, pus_tc=tc, timestamp=CdsShortTimestamp.empty().pack()
        )
        res = verificator.add_tm(srv_1_tm)
        wrapper.log_to_file(srv_1_tm, res)
        # Assert that 4 lines have been written
        with open(self.log_file_name) as file:
            all_lines = file.readlines()
            self.assertEqual(len(all_lines), 4)

    def tearDown(self) -> None:
        log_file = Path(self.log_file_name)
        if log_file.exists():
            os.remove(log_file)



from collections import deque
from unittest import TestCase
from unittest.mock import MagicMock

from spacepackets.ecss import PusTelemetry
from spacepackets.ccsds.time import CdsShortTimestamp
from tmtccmd.tmtc import (
    SpecificApidHandlerBase,
    CcsdsTmHandler,
    GenericApidHandlerBase,
)
from tmtccmd.com import ComInterface
from tmtccmd.tmtc.ccsds_tm_listener import CcsdsTmListener


class ApidHandler(SpecificApidHandlerBase):
    def __init__(self, apid: int):
        super().__init__(apid, None)
        self.was_called = False
        self.called_times = 0
        self.packet_queue = deque()

    def handle_tm(self, packet: bytes, user_args: any):
        if not self.was_called:
            self.was_called = True
        self.called_times += 1
        self.packet_queue.appendleft(packet)


class TestTmHandler(TestCase):
    def setUp(self) -> None:
        self.apid = 0x33

    def test_basic(self):
        tm_handler = ApidHandler(0x01)
        com_if = MagicMock(specs=ComInterface)
        unknown_handler = MagicMock(specs=GenericApidHandlerBase)
        ccsds_handler = CcsdsTmHandler(unknown_handler)
        ccsds_handler.add_apid_handler(tm_handler)
        tm_listener = CcsdsTmListener(tm_handler=ccsds_handler)
        handled_packets = tm_listener.operation(com_if)
        self.assertEqual(handled_packets, 0)
        self.assertTrue(ccsds_handler.has_apid(0x01))
        tm0_raw = PusTelemetry(
            service=1,
            subservice=12,
            apid=0x01,
            timestamp=CdsShortTimestamp.empty().pack(),
        ).pack()
        tm1_raw = PusTelemetry(
            service=5,
            subservice=1,
            apid=0x01,
            timestamp=CdsShortTimestamp.empty().pack(),
        ).pack()
        com_if.receive.return_value = [tm0_raw]
        handled_packets = tm_listener.operation(com_if)
        self.assertEqual(handled_packets, 1)
        self.assertTrue(tm_handler.was_called)
        self.assertEqual(tm_handler.called_times, 1)
        self.assertEqual(tm_handler.packet_queue.pop(), tm0_raw)
        com_if.receive.return_value = [tm0_raw, tm1_raw]
        handled_packets = tm_listener.operation(com_if)
        self.assertEqual(handled_packets, 2)
        self.assertEqual(tm_handler.called_times, 3)
        self.assertEqual(handled_packets, 2)
        self.assertEqual(tm_handler.packet_queue.pop(), tm0_raw)
        self.assertEqual(tm_handler.packet_queue.pop(), tm1_raw)
        unknown_apid = PusTelemetry(
            service=1,
            subservice=12,
            apid=0x02,
            timestamp=CdsShortTimestamp.empty().pack(),
        ).pack()
        com_if.receive.return_value = [unknown_apid]
        handled_packets = tm_listener.operation(com_if)
        self.assertEqual(handled_packets, 1)
        unknown_handler.handle_tm.assert_called_once()



import enum


class Subservice(enum.IntEnum):
    TC_MODE_COMMAND = 1
    TC_MODE_COMMAND_FORCES = 2
    TC_MODE_READ = 3
    TC_MODE_ANNOUNCE = 4
    TC_MODE_ANNOUNCE_RECURSIVE = 5
    TM_MODE_REPLY = 6
    TM_CANT_REACH_MODE = 7
    TM_WRONG_MODE_REPLY = 8



from __future__ import annotations
import json
import logging
import socket
import struct
import enum
from dataclasses import dataclass
from enum import auto
from typing import Tuple

from tmtccmd.util.json import check_json_file, JsonKeyNames

_LOGGER = logging.getLogger(__name__)
DEFAULT_MAX_RECV_SIZE = 1500


@dataclass
class EthAddr:
    ip_addr: str
    port: int

    @property
    def to_tuple(self) -> Tuple[str, int]:
        return self.ip_addr, self.port

    @classmethod
    def from_tuple(cls, addr: Tuple[str, int]) -> EthAddr:
        return cls(addr[0], addr[1])


class TcpIpType(enum.Enum):
    TCP = enum.auto()
    UDP = enum.auto()
    UDP_RECV = enum.auto()


class TcpIpConfigIds(enum.Enum):
    SEND_ADDRESS = auto()
    RECV_ADDRESS = auto()
    RECV_MAX_SIZE = auto()
    # Used by TCP to detect start of space packets
    SPACE_PACKET_ID = auto()


def determine_udp_send_address(json_cfg_path: str) -> EthAddr:
    return determine_tcpip_address(tcpip_type=TcpIpType.UDP, json_cfg_path=json_cfg_path)


def determine_tcp_send_address(json_cfg_path: str) -> EthAddr:
    return determine_tcpip_address(tcpip_type=TcpIpType.TCP, json_cfg_path=json_cfg_path)


def determine_udp_recv_address(json_cfg_path: str) -> EthAddr:
    return determine_tcpip_address(tcpip_type=TcpIpType.UDP_RECV, json_cfg_path=json_cfg_path)


def determine_tcpip_address(tcpip_type: TcpIpType, json_cfg_path: str) -> EthAddr:
    addr = EthAddr("", 0)
    reconfigure_ip_address = False
    if not check_json_file(json_cfg_path=json_cfg_path):
        reconfigure_ip_address = True

    if tcpip_type == TcpIpType.TCP:
        json_key_address = JsonKeyNames.TCPIP_TCP_DEST_IP_ADDRESS.value
        json_key_port = JsonKeyNames.TCPIP_TCP_DEST_PORT.value
        info_string = "TCP destination"
    elif tcpip_type == TcpIpType.UDP:
        json_key_address = JsonKeyNames.TCPIP_UDP_DEST_IP_ADDRESS.value
        json_key_port = JsonKeyNames.TCPIP_UDP_DEST_PORT.value
        info_string = "UDP destination"
    elif tcpip_type == TcpIpType.UDP_RECV:
        json_key_address = JsonKeyNames.TCPIP_UDP_RECV_IP_ADDRESS.value
        json_key_port = JsonKeyNames.TCPIP_UDP_RECV_PORT.value
        info_string = "UDP receive destination"
    else:
        json_key_address = JsonKeyNames.TCPIP_UDP_DEST_IP_ADDRESS.value
        json_key_port = JsonKeyNames.TCPIP_UDP_DEST_PORT.value
        info_string = "UDP destination"

    with open(json_cfg_path, "r") as write:
        load_data = json.load(write)
        if json_key_address not in load_data or json_key_port not in load_data:
            reconfigure_ip_address = True
        else:
            ip_address = load_data[json_key_address]
            port = int(load_data[json_key_port])
            addr = EthAddr(ip_address, port)

    if reconfigure_ip_address:
        addr = prompt_ip_address(type_str=info_string)
        save_to_json = input(f"Do you want to store the {info_string} configuration? ([Y]/n): ")
        if save_to_json.lower() in ["y", "yes", ""]:
            with open(json_cfg_path, "r+") as file:
                json_dict = json.load(file)
                json_dict[json_key_address] = addr.ip_addr
                json_dict[json_key_port] = addr.port
                file.seek(0)
                json.dump(json_dict, file, indent=4)
            _LOGGER.info(f"{info_string} was stored to the JSON file config/tmtcc_config.json.")
            _LOGGER.info("Delete this file or edit it manually to change the used addresses.")
        else:
            _LOGGER.info(f"{info_string} configuration was not stored")
    return addr


def prompt_ip_address(type_str: str) -> EthAddr:
    """Prompt a valid IP address from the user"""
    addr = EthAddr("", 0)
    while True:
        ip_address = input(
            f"Configuring {type_str} IP address. "
            'Please enter the IP address ("localhost" and "any" is valid): '
        )

        check_ip = True
        if ip_address == "localhost":
            ip_address = socket.inet_ntoa(struct.pack("!L", socket.INADDR_LOOPBACK))
            check_ip = False
        elif ip_address == "any":
            ip_address = socket.inet_ntoa(struct.pack("!L", socket.INADDR_ANY))
            check_ip = False
        if check_ip:
            try:
                socket.inet_aton(str(ip_address))
            except socket.error:
                _LOGGER.warning("Invalid IP address format!")
                continue

        port = input(f"Please enter {type_str} port: ")
        addr = EthAddr(ip_address, int(port))

        _LOGGER.info(f"Specified {type_str} IP address: {ip_address}")
        _LOGGER.info(f"Specified {type_str} port: {port}")

        confirm = input("Please confirm selection ([Y]/n): ")
        if confirm.lower() not in ["y", "yes", 1, ""]:
            continue
        break
    return addr


def determine_recv_buffer_len(json_cfg_path: str, tcpip_type: TcpIpType):
    recv_max_size = 0
    reconfigure_recv_buf_size = False
    if tcpip_type == TcpIpType.UDP:
        json_key = JsonKeyNames.TCPIP_UDP_RECV_MAX_SIZE.value
    elif tcpip_type == TcpIpType.TCP:
        json_key = JsonKeyNames.TCPIP_TCP_RECV_MAX_SIZE.value
    else:
        json_key = JsonKeyNames.TCPIP_UDP_RECV_MAX_SIZE.value

    if not check_json_file(json_cfg_path=json_cfg_path):
        reconfigure_recv_buf_size = True
    with open(json_cfg_path, "r") as write:
        load_data = json.load(write)
        if json_key not in load_data:
            reconfigure_recv_buf_size = True
        else:
            recv_max_size = load_data[json_key]
    if reconfigure_recv_buf_size:
        recv_max_size = prompt_recv_buffer_len(tcpip_type=tcpip_type)
        store_size = input("Do you store the maximum receive size configuration? ([Y]/n): ")
        if store_size.lower() in ["y", "yes", "1", ""]:
            with open(json_cfg_path, "r+") as file:
                json_dict = json.load(file)
                json_dict[json_key] = recv_max_size
                file.seek(0)
                json.dump(json_dict, file, indent=4)
    return recv_max_size


def prompt_recv_buffer_len(tcpip_type: TcpIpType) -> int:
    if tcpip_type == TcpIpType.UDP:
        type_str = "UDP"
    else:
        type_str = "TCP"
    while True:
        recv_max_size = input(
            f"Please enter maximum receive size for {type_str} packets [1500 default]: "
        )
        if not recv_max_size:
            return DEFAULT_MAX_RECV_SIZE
        if not recv_max_size.isdigit():
            _LOGGER.warning("Specified size is not a number.")
            continue
        else:
            try:
                recv_max_size = int(recv_max_size)
            except ValueError:
                _LOGGER.warning("Specified input invalid")
                continue
            break
    return recv_max_size



"""Please note that this module does not contain configuration helpers, for example
to convert CLI or GUI parameters into the internalized CFDP classes. You can find all those
helpers inside the :py:mod:`tmtccmd.config.cfdp` module."""

from .request import CfdpRequestWrapper
from spacepackets.cfdp import TransactionId



from __future__ import annotations

from typing import Mapping, Union, Optional

from spacepackets.util import UnsignedByteField


class ComponentIdBase(UnsignedByteField):
    """Base class for unsigned object IDs with different byte sizes"""

    def __init__(self, obj_id: int, byte_len: int, name: Optional[str] = None):
        super().__init__(val=obj_id, byte_len=byte_len)
        if name is None:
            self.name = "Unknown"
        else:
            self.name = name

    def __str__(self):
        return f"Object ID {self.name} with ID {self.as_hex_string}"

    @property
    def as_hex_string(self) -> str:
        if self.byte_len == 1:
            return f"{self.obj_id:#04x}"
        elif self.byte_len == 2:
            return f"{self.obj_id:#06x}"
        elif self.byte_len == 4:
            return f"{self.obj_id:#010x}"
        elif self.byte_len == 8:
            return f"{self.obj_id:#018x}"
        return f"{self.obj_id:#x}"

    @property
    def obj_id(self) -> int:
        return int(self)

    @obj_id.setter
    def obj_id(self, obj_id: Union[int, bytes]):
        """This setter function takes a raw byte stream to deserialize an object ID, or the ID
        as an integer.

        :raise ValueError: Invalid ID
        """
        self.value = obj_id


class ComponentIdU32(ComponentIdBase):
    """A helper object for a unique object identifier which has a raw unsigned
    32-bit representation.

    >>> obj_id = ComponentIdU32(42, "Object with the answer to everything")
    >>> int(obj_id)
    42
    >>> obj_id.name
    'Object with the answer to everything'
    >>> obj_id.as_bytes.hex(sep=",")
    '00,00,00,2a'
    >>> obj_id.as_hex_string
    '0x0000002a'
    """

    def __init__(self, obj_id: int, name: Optional[str] = None):
        super().__init__(obj_id, 4, name)

    def __repr__(self):
        return f"{self.__class__.__name__}(object_id={self.obj_id!r}, name={self.name!r})"

    @classmethod
    def from_bytes_typed(cls, obj_id_as_bytes: bytes) -> ComponentIdU32:
        obj_id = ComponentIdU32(obj_id=0)
        obj_id.obj_id = obj_id_as_bytes
        return obj_id


class ComponentIdU16(ComponentIdBase):
    """A helper object for a unique object identifier which has a raw unsigned
    16-bit representation.
    """

    def __init__(self, obj_id: int, name: Optional[str] = None):
        super().__init__(obj_id, 2, name)

    def __repr__(self):
        return f"{self.__class__.__name__}(object_id={self.obj_id!r}, name={self.name!r})"

    @classmethod
    def from_bytes_typed(cls, obj_id_as_bytes: bytes) -> ComponentIdU16:
        obj_id = ComponentIdU16(obj_id=0)
        obj_id.obj_id = obj_id_as_bytes
        return obj_id


class ComponentIdU8(ComponentIdBase):
    """A helper object for a unique object identifier which has a raw unsigned
    8-bit representation.
    """

    def __init__(self, obj_id: int, name: Optional[str] = None):
        super().__init__(obj_id, 1, name)

    def __repr__(self):
        return f"{self.__class__.__name__}(object_id={self.obj_id!r}, name={self.name!r})"

    @classmethod
    def from_bytes_typed(cls, obj_id_as_bytes: bytes) -> ComponentIdU8:
        obj_id = ComponentIdU8(obj_id=0)
        obj_id.obj_id = obj_id_as_bytes
        return obj_id


ComponentIdMapping = Mapping[bytes, ComponentIdBase]


ObjectIdBase = ComponentIdBase
"""Deprecated type defintion for :py:class:`ComponentIdBase`"""

ObjectIdU32 = ComponentIdU32
"""Deprecated type defintion for :py:class:`ComponentIdU32`"""

ObjectIdU16 = ComponentIdU16
"""Deprecated type defintion for :py:class:`ComponentIdU16`"""

ObjectIdU8 = ComponentIdU8
"""Deprecated type defintion for :py:class:`ComponentIdU8`"""

ObjectIdMapping = ComponentIdMapping
"""Deprecated type defintion for :py:class:`ComponentIdMapping`"""

ObjectIdDictT = ObjectIdMapping
"""Deprecated type defintion for :py:class:`ObjectIdMapping`"""



import time
from typing import cast
from unittest import TestCase
from datetime import timedelta
from unittest.mock import MagicMock, ANY

from spacepackets.ecss import PusTelecommand
from tmtccmd.com import ComInterface
from tmtccmd.tmtc.ccsds_seq_sender import SequentialCcsdsSender, SenderMode
from tmtccmd.tmtc.handler import TcHandlerBase, SendCbParams
from tmtccmd.tmtc.procedure import TreeCommandingProcedure
from tmtccmd.tmtc.queue import QueueWrapper, DefaultPusQueueHelper


class TestSendReceive(TestCase):
    def setUp(self) -> None:
        self.apid = 0x22
        self.queue_wrapper = QueueWrapper.empty()
        self.queue_helper = DefaultPusQueueHelper(
            self.queue_wrapper,
            tc_sched_timestamp_len=4,
            pus_verificator=None,
            default_pus_apid=None,
            seq_cnt_provider=None,
        )
        self.tc_handler_mock = MagicMock(spec=TcHandlerBase)
        self.com_if = MagicMock(spec=ComInterface)
        self.seq_sender = SequentialCcsdsSender(self.queue_wrapper, self.tc_handler_mock)

    def test_basic(self):
        res = self.seq_sender.operation(self.com_if)
        # Queue is empty initially
        self.assertEqual(res.mode, SenderMode.DONE)
        self.assertEqual(self.seq_sender.mode, SenderMode.DONE)
        self.assertTrue(self.seq_sender.no_delay_remaining())
        self.queue_helper.add_raw_tc(bytes([0, 1, 2]))
        # One queue entry which should be handled immediately
        self.seq_sender.queue_wrapper = self.queue_wrapper
        self.assertEqual(self.seq_sender.mode, SenderMode.BUSY)
        # Is busy now, so does not accept new queue unless forced
        with self.assertRaises(ValueError):
            self.seq_sender.queue_wrapper = self.queue_wrapper
        self.seq_sender.operation(self.com_if)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.info.procedure, TreeCommandingProcedure.empty())
        raw_tc_entry = send_cb_params.entry.to_raw_tc_entry()
        self.assertEqual(raw_tc_entry.tc, bytes([0, 1, 2]))
        # Queue should be empty now
        # Called twice for each operation call
        self.assertEqual(self.tc_handler_mock.queue_finished_cb.call_count, 2)
        self.assertFalse(self.queue_wrapper.queue)
        self.assertEqual(self.seq_sender.mode, SenderMode.DONE)
        self.queue_helper.add_raw_tc(bytes([3, 2, 1]))
        self.seq_sender.resume()
        self.assertEqual(self.seq_sender.mode, SenderMode.BUSY)
        res = self.seq_sender.operation(self.com_if)
        self.assertTrue(res.tc_sent)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.entry.to_raw_tc_entry().tc, bytes([3, 2, 1]))

    def test_with_wait_entry(self):
        wait_delay = 0.01
        self.queue_helper.add_raw_tc(bytes([3, 2, 1]))
        self.queue_helper.add_wait(timedelta(seconds=wait_delay))
        self.queue_helper.add_raw_tc(bytes([1, 2, 3]))
        # Resume call necessary
        self.assertEqual(self.seq_sender.mode, SenderMode.DONE)
        self.seq_sender.resume()
        res = self.seq_sender.operation(self.com_if)
        self.assertEqual(res.mode, SenderMode.BUSY)
        self.assertTrue(res.tc_sent)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.entry.to_raw_tc_entry().tc, bytes([3, 2, 1]))
        self.assertTrue(self.seq_sender.no_delay_remaining())
        # 2 queue entries remaining
        self.assertEqual(len(self.queue_helper.queue_wrapper.queue), 2)
        # Now the wait entry should be handled
        res = self.seq_sender.operation(self.com_if)
        self.assertFalse(self.seq_sender.no_delay_remaining())
        self.assertFalse(res.tc_sent)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(
            send_cb_params.entry.to_wait_entry().wait_time,
            timedelta(seconds=wait_delay),
        )
        # Now no TCs should be sent for 10 ms
        self.assertEqual(len(self.queue_helper.queue_wrapper.queue), 1)
        self.assertEqual(res.mode, SenderMode.BUSY)
        res = self.seq_sender.operation(self.com_if)
        self.assertFalse(res.tc_sent)
        self.assertEqual(len(self.queue_helper.queue_wrapper.queue), 1)
        # After a delay, TC should be sent
        time.sleep(wait_delay)
        res = self.seq_sender.operation(self.com_if)
        self.assertTrue(res.tc_sent)
        self.assertEqual(len(self.queue_helper.queue_wrapper.queue), 0)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.entry.to_raw_tc_entry().tc, bytes([1, 2, 3]))

    def test_interpacket_delay(self):
        delay_ms = 20
        inter_packet_delay = timedelta(milliseconds=delay_ms)
        ping_cmd = PusTelecommand(apid=self.apid, service=17, subservice=1)
        self.queue_helper.add_pus_tc(ping_cmd)
        self.queue_helper.add_packet_delay_ms(delay_ms)
        self.queue_helper.add_ccsds_tc(ping_cmd.to_space_packet())
        self.queue_helper.add_raw_tc(bytes([0, 1, 2]))
        # Send first TC, assert delay of 10 ms, then send last packet
        res = self.seq_sender.operation(self.com_if)
        self.assertEqual(res.longest_rem_delay, timedelta())
        self.assertTrue(res.tc_sent)
        self.assertEqual(res.next_entry_is_tc, False)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.com_if, self.com_if)
        self.assertIsNone(send_cb_params.info.procedure)
        self.assertEqual(send_cb_params.entry.to_pus_tc_entry().pus_tc, ping_cmd)
        res = self.seq_sender.operation(self.com_if)
        self.assertFalse(res.tc_sent)
        self.assertEqual(res.next_entry_is_tc, True)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertIsNone(send_cb_params.info.procedure)
        self.assertEqual(
            send_cb_params.entry.to_packet_delay_entry().delay_time.total_seconds(),
            inter_packet_delay.total_seconds(),
        )
        self.assertTrue(inter_packet_delay * 0.8 <= res.longest_rem_delay <= inter_packet_delay)
        res = self.seq_sender.operation(self.com_if)
        # No TC sent
        self.assertFalse(res.tc_sent)
        self.assertEqual(res.next_entry_is_tc, True)
        self.assertEqual(len(self.queue_wrapper.queue), 2)
        time.sleep(inter_packet_delay.total_seconds())
        res = self.seq_sender.operation(self.com_if)
        # TC sent
        self.assertTrue(res.tc_sent)
        self.assertEqual(len(self.queue_wrapper.queue), 1)
        res = self.seq_sender.operation(self.com_if)
        # No TC sent, delay after each packet
        self.assertFalse(res.tc_sent)
        self.assertEqual(len(self.queue_wrapper.queue), 1)
        self.assertTrue(0.8 * inter_packet_delay < res.longest_rem_delay <= inter_packet_delay)
        # Delay 10 ms
        time.sleep(inter_packet_delay.total_seconds())
        res = self.seq_sender.operation(self.com_if)
        self.assertTrue(res.tc_sent)
        # Queue is empty now, but this should still be set to False
        self.assertEqual(res.next_entry_is_tc, False)
        self.tc_handler_mock.send_cb.assert_called_with(ANY)
        call_args = self.tc_handler_mock.send_cb.call_args
        send_cb_params = cast(SendCbParams, call_args.args[0])
        self.assertEqual(send_cb_params.entry.to_raw_tc_entry().tc, bytes([0, 1, 2]))

    def test_delay_at_end(self):
        delay_at_end = timedelta(milliseconds=50)
        self.queue_helper.add_raw_tc(bytes([3, 2, 1]))
        self.queue_helper.add_wait(delay_at_end)
        self.seq_sender.resume()
        res = self.seq_sender.operation(self.com_if)
        self.assertTrue(res.tc_sent)
        self.assertEqual(res.longest_rem_delay, timedelta())
        res = self.seq_sender.operation(self.com_if)
        self.assertFalse(res.tc_sent)
        self.assertFalse(self.seq_sender.no_delay_remaining())
        self.assertTrue(0.8 * delay_at_end < res.longest_rem_delay <= delay_at_end)
        self.assertEqual(self.seq_sender.mode, SenderMode.BUSY)
        time.sleep(delay_at_end.total_seconds())
        self.assertTrue(self.seq_sender.no_delay_remaining())
        self.seq_sender.operation(self.com_if)
        self.assertEqual(self.seq_sender.mode, SenderMode.DONE)



from typing import Mapping


class RetvalInfo:
    def __init__(self):
        self.id: int = 0
        self.name: str = ""
        self.info: str = ""
        self.if_name: str = ""

    def id_as_hex(self) -> str:
        return f"{self.id:#04x}"

    @property
    def subsystem_id(self) -> int:
        return (self.id >> 8) & 0xFF

    @property
    def unique_id(self) -> int:
        return self.id & 0xFF


RetvalMapping = Mapping[int, RetvalInfo]

RetvalDictT = RetvalMapping
"""Deprecatd typedef for :py:class:`RetvalMapping`."""



import collections
import logging
import threading
from typing import List, Optional

from tmtccmd.com import ComInterface, ReceptionDecodeError
from tmtccmd.com.serial_base import SerialComBase, SerialCfg, SerialCommunicationType
from cobs import cobs


class SerialCobsComIF(SerialComBase, ComInterface):
    """Serial communication interface which uses the
    `COBS protocol <https://pypi.org/project/cobs/>`_ to encode and decode packets.

    This class will spin up a receiver thread on the :meth:`open` call to poll
    for COBS encoded packets.
    This means that the :meth:`close` call might block until the receiver thread has shut down.
    """

    def __init__(self, ser_cfg: SerialCfg):
        super().__init__(
            logging.getLogger(__name__),
            ser_cfg=ser_cfg,
            ser_com_type=SerialCommunicationType.COBS,
        )
        self.__polling_shutdown = threading.Event()
        self.__reception_thread: Optional[threading.Thread] = None
        self.__reception_buffer = collections.deque()

    @property
    def id(self) -> str:
        return self.ser_cfg.com_if_id

    def initialize(self, args: any = None) -> any:
        pass

    def open(self, args: any = None) -> None:
        """Spins up a receiver thread to permanently check for new COBS encoded packets."""
        super().open_port()
        self.__polling_shutdown.clear()
        self.__reception_thread = threading.Thread(target=self.__poll_cobs_packets, daemon=True)
        self.__reception_thread.start()

    def is_open(self) -> bool:
        return self.serial is not None

    def close(self, args: any = None) -> None:
        self.__polling_shutdown.set()
        self.__reception_thread.join(0.4)
        super().close_port()

    def send(self, data: bytes):
        encoded = bytearray([0])
        encoded.extend(cobs.encode(data))
        encoded.append(0)
        self.serial.write(encoded)

    def receive(self, parameters: any = 0) -> List[bytes]:
        packet_list = []
        while self.__reception_buffer:
            data = self.__reception_buffer.pop()
            try:
                packet_list.append(cobs.decode(data))
            except cobs.DecodeError as e:
                raise ReceptionDecodeError(f"COBS decoding error: {e}", e)
        return packet_list

    def data_available(self, timeout: float, parameters: any = 0) -> int:
        return SerialComBase.data_available_from_queue(timeout, self.__reception_buffer)

    def __poll_cobs_packets(self):
        last_byte_was_zero = False
        # Poll permanently, but it is possible to join this thread every 200 ms
        self.serial.timeout = 0.2
        while True:
            byte = self.serial.read()
            if len(byte) == 1:
                if byte[0] == 0:
                    last_byte_was_zero = True
                else:
                    if last_byte_was_zero:
                        self.serial.timeout = 0.1
                        possible_cobs_frame = bytearray(byte)
                        possible_cobs_frame.extend(self.serial.read_until(bytes([0])))
                        self.__reception_buffer.appendleft(possible_cobs_frame[:-1])
                        self.serial.timeout = 0.2
                        last_byte_was_zero = True
                    else:
                        broken_cobs_frame = self.serial.read_until(bytes([0]))
                        self.logger.warning(
                            "Discarding possibly broken COBS frame: "
                            f"{broken_cobs_frame.hex(sep=',')}"
                        )
                        last_byte_was_zero = True
            elif self.__polling_shutdown.is_set():
                break



import os
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from tmtccmd.com.utils import determine_com_if


class TestUtils(TestCase):
    def setUp(self) -> None:
        self.json_file = "test.json"

    def test_com_if_utils(self):
        with patch("tmtccmd.com.utils.wrapped_prompt", side_effect=["0", "yes"]):
            test_dict = {"test-com-if": ("Some more info", None)}
            com_if = determine_com_if(test_dict, self.json_file, True)
            self.assertEqual(com_if, "test-com-if")
            with open(self.json_file) as file:
                lines = file.readlines()
                lines[0] = "{\n"
                lines[1] = '    "com_if": "test-com-if"\n'
                lines[2] = "}"
            os.remove(self.json_file)
        with patch("tmtccmd.com.utils.wrapped_prompt", side_effect=["0", "no"]):
            test_dict = {"test-com-if": ("Some more info", None)}
            com_if = determine_com_if(test_dict, self.json_file, True)
            self.assertEqual(com_if, "test-com-if")
            with open(self.json_file) as file:
                lines = file.readlines()
                lines[0] = "{}"
        with patch(
            "tmtccmd.com.utils.wrapped_prompt",
            side_effect=["1", "0", "no"],
        ):
            test_dict = {"test-com-if": ("Some more info", None)}
            com_if = determine_com_if(test_dict, self.json_file, True)
            self.assertEqual(com_if, "test-com-if")
        with patch(
            "tmtccmd.com.utils.wrapped_prompt",
            side_effect=["blub", "0", "no"],
        ):
            test_dict = {"test-com-if": ("Some more info", None)}
            com_if = determine_com_if(test_dict, self.json_file, True)
            self.assertEqual(com_if, "test-com-if")

    def tearDown(self) -> None:
        path = Path(self.json_file)
        if path.exists():
            os.remove(path)



from unittest import TestCase

from spacepackets.ecss import PusService
from tmtccmd.pus.s20_fsfw_param_defs import (
    create_scalar_boolean_parameter,
    CustomSubservice,
    Parameter,
)
from tmtccmd.pus.s20_fsfw_param import Service20FsfwTm, Service20ParamDumpWrapper


class TestSrv20Tm(TestCase):
    def setUp(self):
        self.obj_id = bytes([0x01, 0x02, 0x03, 0x04])
        self.boolean_param = create_scalar_boolean_parameter(
            object_id=self.obj_id, domain_id=1, unique_id=5, parameter=True
        )
        self.tm = Service20FsfwTm(
            subservice=CustomSubservice.TM_DUMP_REPLY,
            timestamp=bytes(),
            source_data=self.boolean_param.pack(),
        )

    def test_state(self):
        self.assertEqual(self.tm.timestamp, bytes())
        self.assertEqual(self.tm.service, PusService.S20_PARAMETER)
        self.assertEqual(self.tm.subservice, CustomSubservice.TM_DUMP_REPLY)
        self.assertEqual(self.tm.source_data, self.boolean_param.pack())
        self.assertEqual(self.tm.object_id, self.obj_id)

    def test_unpack(self):
        tm_raw = self.tm.pack()
        tm_unpacked = Service20FsfwTm.unpack(raw_telemetry=tm_raw, timestamp_len=0)
        self.assertEqual(self.tm, tm_unpacked)
        param_unpacked = Parameter.unpack(tm_unpacked.source_data)
        self.assertEqual(param_unpacked, self.boolean_param)

    def test_dump_wrapper(self):
        self.dump_wrapper = Service20ParamDumpWrapper(param_tm=self.tm)
        param = self.dump_wrapper.get_param()
        self.assertEqual(param, self.boolean_param)



from dataclasses import dataclass
import enum
import threading
from typing import Any, Optional

from tmtccmd import CcsdsTmtcBackend
from tmtccmd.config import CoreComInterfaces

CONNECT_BTTN_STYLE = (
    "background-color: #1fc600;"
    "border-style: inset;"
    "font: bold;"
    "color: black;"
    "padding: 6px;"
    "border-width: 2px;"
    "border-radius: 6px;"
)


DISCONNECT_BTTN_STYLE = (
    "background-color: orange;"
    "border-style: inset;"
    "font: bold;"
    "color: black;"
    "padding: 6px;"
    "border-width: 2px;"
    "border-radius: 6px;"
)


COMMAND_BUTTON_STYLE = (
    "background-color: #cdeefd;"
    "border-style: inset;"
    "font: bold;"
    "color: black;"
    "padding: 6px;"
    "border-width: 2px;"
    "border-radius: 6px;"
)


class WorkerOperationsCode(enum.IntEnum):
    OPEN_COM_IF = 0
    CLOSE_COM_IF = 1
    ONE_QUEUE_MODE = 2
    LISTEN_FOR_TM = 3
    UPDATE_BACKEND_MODE = 4
    IDLE = 5


class ComIfRefCount:
    def __init__(self):
        self.lock = threading.Lock()
        self.com_if_used = False
        self.user_cnt = 0

    def add_user(self):
        with self.lock:
            self.user_cnt += 1

    def remove_user(self):
        with self.lock:
            if self.user_cnt > 0:
                self.user_cnt -= 1

    def is_used(self):
        with self.lock:
            if self.user_cnt > 0:
                return True
            return False


class LocalArgs:
    def __init__(self, op_code: WorkerOperationsCode, op_code_args: Any = None):
        self.op_code = op_code
        self.op_args = op_code_args


class SharedArgs:
    def __init__(self, backend: CcsdsTmtcBackend):
        self.state_lock = threading.Lock()
        self.com_if_ref_tracker = ComIfRefCount()
        self.tc_lock = threading.Lock()
        self.backend = backend


@dataclass
class FrontendState:
    current_com_if = CoreComInterfaces.UNSPECIFIED.value
    current_cmd_path: Optional[str] = None
    auto_connect_tm_listener = True
    last_com_if = CoreComInterfaces.UNSPECIFIED.value
    current_com_if_key = CoreComInterfaces.UNSPECIFIED.value
    print_tm = False
    print_raw_tm = False



from typing import Optional, cast

from cfdppy.defs import CfdpRequestType
from tmtccmd.config.defs import CfdpParams


class CfdpRequestBase:
    def __init__(self, req_type: CfdpRequestType):
        self.req_type = req_type


class PutRequestCfgWrapper(CfdpRequestBase):
    def __init__(self, cfg: CfdpParams):
        super().__init__(CfdpRequestType.PUT)
        self.cfg = cfg

    def __repr__(self):
        return f"{self.__class__.__name__}(cfg={self.cfg})"


class CfdpRequestWrapper:
    def __init__(self, base: Optional[CfdpRequestBase]):
        self.base = base

    @property
    def request_type(self) -> CfdpRequestType:
        return self.base.req_type

    @property
    def request(self) -> CfdpRequestType:
        return self.base.req_type

    def to_put_request(self) -> PutRequestCfgWrapper:
        if self.base.req_type != CfdpRequestType.PUT:
            raise TypeError(f"Request is not a {PutRequestCfgWrapper.__name__}: {self.base!r}")
        return cast(PutRequestCfgWrapper, self.base)



import logging
import sys
from typing import Optional, cast, Sequence

from spacepackets.ccsds import PacketId
from tmtccmd.config.defs import CoreComInterfaces
from tmtccmd.com import ComInterface
from tmtccmd.com.serial_base import (
    SerialCommunicationType,
    SerialCfg,
)

from tmtccmd.com.serial_dle import SerialDleComIF
from tmtccmd.com.serial_cobs import SerialCobsComIF

from tmtccmd.com.ser_utils import determine_com_port, determine_baud_rate
from tmtccmd.com.tcpip_utils import TcpIpType, EthAddr
from tmtccmd.com.udp import UdpClient
from tmtccmd.com.tcp import TcpSpacepacketsClient

_LOGGER = logging.getLogger(__name__)


class ComCfgBase:
    def __init__(
        self,
        com_if_key: str,
        json_cfg_path: str,
    ):
        self.com_if_key = com_if_key
        self.json_cfg_path = json_cfg_path


class TcpipCfg(ComCfgBase):
    def __init__(
        self,
        if_type: TcpIpType,
        com_if_key: str,
        json_cfg_path: str,
        send_addr: EthAddr,
        space_packet_ids: Optional[Sequence[PacketId]],
        recv_addr: Optional[EthAddr] = None,
    ):
        super().__init__(com_if_key, json_cfg_path)
        self.space_packet_ids = space_packet_ids
        self.if_type = if_type
        self.send_addr = send_addr
        self.recv_addr = recv_addr


class SerialCfgWrapper(ComCfgBase):
    def __init__(self, com_if_key: str, json_cfg_path: str, serial_cfg: SerialCfg):
        super().__init__(com_if_key=com_if_key, json_cfg_path=json_cfg_path)
        self.serial_cfg = serial_cfg


def create_com_interface_cfg_default(
    com_if_key: str, json_cfg_path: str, space_packet_ids: Optional[Sequence[PacketId]]
) -> Optional[ComCfgBase]:
    if com_if_key == CoreComInterfaces.DUMMY.value:
        return ComCfgBase(com_if_key=com_if_key, json_cfg_path=json_cfg_path)
    elif com_if_key == CoreComInterfaces.UDP.value:
        return default_tcpip_cfg_setup(
            com_if_key=com_if_key,
            json_cfg_path=json_cfg_path,
            tcpip_type=TcpIpType.UDP,
            space_packet_ids=space_packet_ids,
        )
    elif com_if_key == CoreComInterfaces.TCP.value:
        assert space_packet_ids is not None
        return default_tcpip_cfg_setup(
            com_if_key=com_if_key,
            json_cfg_path=json_cfg_path,
            tcpip_type=TcpIpType.TCP,
            space_packet_ids=space_packet_ids,
        )
    elif com_if_key in [
        CoreComInterfaces.SERIAL_DLE.value,
        CoreComInterfaces.SERIAL_COBS.value,
        CoreComInterfaces.SERIAL_QEMU.value,
    ]:
        # For a serial communication interface, there are some configuration values like
        # baud rate and serial port which need to be set once but are expected to stay
        # the same for a given machine. Therefore, we use a JSON file to store and extract
        # those values
        cfg = SerialCfg(baud_rate=0, com_if_id=com_if_key, serial_timeout=0, serial_port="")
        default_serial_cfg_baud_and_port_setup(json_cfg_path, cfg)
        return SerialCfgWrapper(com_if_key=com_if_key, json_cfg_path=json_cfg_path, serial_cfg=cfg)
    else:
        return None


def create_com_interface_default(cfg: ComCfgBase) -> Optional[ComInterface]:
    """Return the desired communication interface object

    :param cfg: Generic configuration
    :return:
    """
    if cfg is None:
        raise ValueError("Passed ComIF configuration is empty")
    if cfg.com_if_key == "":
        _LOGGER.warning("COM Interface key string is empty. Using dummy COM interface")
    try:
        return __create_com_if(cfg)
    except ConnectionRefusedError:
        _LOGGER.exception("TCP/IP connection refused")
        if cfg.com_if_key == CoreComInterfaces.UDP.value:
            _LOGGER.warning("Make sure that a UDP server is running")
        if cfg.com_if_key == CoreComInterfaces.TCP.value:
            _LOGGER.warning("Make sure that a TCP server is running")
        sys.exit(1)
    except (IOError, OSError):
        _LOGGER.exception("Error setting up communication interface")
        sys.exit(1)


def __create_com_if(cfg: ComCfgBase) -> Optional[ComInterface]:
    from tmtccmd.com.dummy import DummyComIF
    from tmtccmd.com.qemu import QEMUComIF

    if (
        cfg.com_if_key == CoreComInterfaces.UDP.value
        or cfg.com_if_key == CoreComInterfaces.TCP.value
    ):
        communication_interface = create_default_tcpip_interface(cast(TcpipCfg, cfg))
    elif cfg.com_if_key in [
        CoreComInterfaces.SERIAL_DLE.value,
        CoreComInterfaces.SERIAL_COBS.value,
    ]:
        serial_cfg_wrapper = cast(SerialCfgWrapper, cfg)
        communication_interface = create_default_serial_interface(
            com_if_key=cfg.com_if_key,
            json_cfg_path=cfg.json_cfg_path,
            serial_cfg=serial_cfg_wrapper.serial_cfg,
        )
    elif cfg.com_if_key == CoreComInterfaces.SERIAL_QEMU.value:
        serial_cfg_wrapper = cast(SerialCfgWrapper, cfg)
        # TODO: Move to new model where config is passed externally
        communication_interface = QEMUComIF(
            serial_cfg=serial_cfg_wrapper.serial_cfg,
            ser_com_type=SerialCommunicationType.DLE_ENCODING,
        )
    else:
        communication_interface = DummyComIF()
    if communication_interface is None:
        _LOGGER.warning("Invalid communication interface, is None")
        return communication_interface
    communication_interface.initialize()
    return communication_interface


def default_tcpip_cfg_setup(
    com_if_key: str,
    tcpip_type: TcpIpType,
    json_cfg_path: str,
    space_packet_ids: Optional[Sequence[PacketId]],
) -> TcpipCfg:
    """Default setup for TCP/IP communication interfaces. This intantiates all required data in the
    globals manager so a TCP/IP communication interface can be built with
    :func:`create_default_tcpip_interface`

    :param com_if_key:
    :param tcpip_type:
    :param json_cfg_path:
    :param space_packet_ids:       Required if the TCP com interface needs to parse space packets
    :return:
    """
    from tmtccmd.com.tcpip_utils import (
        determine_udp_send_address,
        determine_tcp_send_address,
    )

    if tcpip_type == TcpIpType.UDP:
        send_addr = determine_udp_send_address(json_cfg_path=json_cfg_path)
    elif tcpip_type == TcpIpType.TCP:
        send_addr = determine_tcp_send_address(json_cfg_path=json_cfg_path)
    else:
        raise ValueError("Invalid TCP/IP server type")
    cfg = TcpipCfg(
        com_if_key=com_if_key,
        if_type=tcpip_type,
        json_cfg_path=json_cfg_path,
        send_addr=send_addr,
        space_packet_ids=space_packet_ids,
    )
    return cfg


def default_serial_cfg_baud_and_port_setup(json_cfg_path: str, cfg: SerialCfg):
    """Default setup for serial interfaces.

    :param json_cfg_path:
    :param cfg: The baud and serial port parameter will be set in this dataclass
    :return:
    """
    baud_rate = determine_baud_rate(json_cfg_path=json_cfg_path)
    serial_port = determine_com_port(json_cfg_path=json_cfg_path)
    cfg.serial_port = serial_port
    cfg.baud_rate = baud_rate


def create_default_tcpip_interface(tcpip_cfg: TcpipCfg) -> Optional[ComInterface]:
    """Create a default serial interface. Requires a certain set of global variables set up. See
    :py:func:`default_tcpip_cfg_setup` for more details.

    :param tcpip_cfg: Configuration parameters
    :return:
    """
    communication_interface = None
    if tcpip_cfg.com_if_key == CoreComInterfaces.UDP.value:
        communication_interface = UdpClient(
            com_if_id=tcpip_cfg.com_if_key,
            send_address=tcpip_cfg.send_addr,
            recv_addr=tcpip_cfg.recv_addr,
        )
    elif tcpip_cfg.com_if_key == CoreComInterfaces.TCP.value:
        assert tcpip_cfg.space_packet_ids is not None
        communication_interface = TcpSpacepacketsClient(
            com_if_id=tcpip_cfg.com_if_key,
            space_packet_ids=tcpip_cfg.space_packet_ids,
            inner_thread_delay=0.5,
            target_address=tcpip_cfg.send_addr,
        )
    return communication_interface


def create_default_serial_interface(
    com_if_key: str, json_cfg_path: str, serial_cfg: SerialCfg
) -> Optional[ComInterface]:
    """Create a default serial interface. Requires a certain set of global variables set up. See
    :func:`set_up_serial_cfg` for more details.

    :param com_if_key:
    :param json_cfg_path:
    :param serial_cfg: Generic serial configuration parameters
    :return:
    """
    try:
        if com_if_key == CoreComInterfaces.SERIAL_DLE.value:
            # Ignore the DLE config for now, it is not that important anyway
            communication_interface = SerialDleComIF(ser_cfg=serial_cfg, dle_cfg=None)
        elif com_if_key == CoreComInterfaces.SERIAL_COBS.value:
            communication_interface = SerialCobsComIF(ser_cfg=serial_cfg)
        else:
            # TODO: Maybe print valid keys?
            _LOGGER.warning(f"Invalid COM IF key {com_if_key} for a serial interface")
            return None
    except KeyError as e:
        _LOGGER.warning("Serial configuration global not configured properly")
        raise e
    return communication_interface



from __future__ import annotations
import dataclasses
import enum
import struct
from typing import Union, Sequence

from spacepackets.ecss import Ptc, PfcUnsigned, PfcSigned, PfcReal


class CustomSubservice(enum.IntEnum):
    TC_LOAD = 128
    TC_DUMP = 129
    TM_DUMP_REPLY = 130


@dataclasses.dataclass
class ParameterId:
    domain_id: int
    unique_id: int
    linear_index: int

    @classmethod
    def empty(cls) -> ParameterId:
        return cls(0, 0, 0)

    def as_u32(self) -> int:
        return (self.domain_id << 24) | (self.unique_id << 16) | self.linear_index

    def pack(self) -> bytes:
        if self.domain_id > pow(2, 8) - 1:
            raise ValueError(f"domain ID {self.domain_id} too large, max value is {pow(2, 8) - 1}")
        if self.unique_id > pow(2, 8) - 1:
            raise ValueError(f"unique ID {self.unique_id} too large, max value is {pow(2, 8) - 1}")
        raw = bytearray([self.domain_id, self.unique_id])
        raw.extend(struct.pack("!H", self.linear_index))
        return raw

    @classmethod
    def unpack(cls, data: bytes) -> ParameterId:
        if len(data) < 4:
            raise ValueError("raw parameter ID must be at least 4 bytes wide")
        domain_id = data[0]
        unique_id = data[1]
        linear_index = struct.unpack("!H", data[2:4])[0]
        return cls(domain_id, unique_id, linear_index)


@dataclasses.dataclass
class FsfwParamId:
    """Wrapper for the whole FSFW specific parameter data. It contains the ECSS PTC and PFC numbers
    and the number of columns and rows in the parameter.
    See https://ecss.nl/standard/ecss-e-st-70-41c-space-engineering-telemetry-and-telecommand-packet-utilization-15-april-2016/
    p.428 for more information.

    :param ptc:     ECSS PTC number
    :param pfc:     ECSS PFC number
    :param rows:     Number of rows in parameter (for matrix entries, 1 for vector entries,
        1 for scalar entries)
    :param columns:  Number of columns in parameter (for matrix or vector entries,
        1 for scalar entries)
    :return: Parameter information field as 4 byte bytearray
    """  # noqa: E501

    object_id: bytes
    param_id: ParameterId
    ptc: Ptc
    pfc: int
    rows: int
    columns: int

    @classmethod
    def unpack(cls, data: bytes) -> FsfwParamId:
        if len(data) < 12:
            raise ValueError("passed raw parameter data size smaller than 12 bytes")
        try:
            ptc = Ptc(data[8])
        except TypeError:
            raise ValueError(f"ptc with unknown raw value {data[8]}")
        return cls(
            object_id=data[0:4],
            param_id=ParameterId.unpack(data[4:8]),
            ptc=ptc,
            pfc=data[9],
            rows=data[10],
            columns=data[11],
        )

    def pack(self) -> bytearray:
        """Convert the wrapper to the raw byte format expected for PUS TC or PUS TM creation."""
        raw = bytearray(self.object_id)
        raw.extend(self.param_id.pack())
        raw.append(self.ptc)
        raw.append(self.pfc)
        raw.append(self.rows)
        raw.append(self.columns)
        return raw


@dataclasses.dataclass
class Parameter:
    fsfw_param_id: FsfwParamId
    param_raw: bytes

    @property
    def ptc(self):
        return self.fsfw_param_id.ptc

    @property
    def pfc(self):
        return self.fsfw_param_id.pfc

    @property
    def object_id(self):
        return self.fsfw_param_id.object_id

    @property
    def rows(self):
        return self.fsfw_param_id.rows

    @property
    def columns(self):
        return self.fsfw_param_id.columns

    @property
    def param_id(self):
        return self.fsfw_param_id.param_id

    @classmethod
    def empty(cls):
        return cls(
            fsfw_param_id=FsfwParamId(
                object_id=bytes([0, 0, 0, 0]),
                param_id=ParameterId.empty(),
                ptc=Ptc.DEDUCED,
                pfc=0,
                rows=0,
                columns=0,
            ),
            param_raw=bytes(),
        )

    def pack(self) -> bytearray:
        """Convert the wrapper to the raw byte format expected for PUS TC or PUS TM creation."""
        raw = self.fsfw_param_id.pack()
        raw.extend(self.param_raw)
        return raw

    @classmethod
    def unpack(cls, data: bytes) -> Parameter:
        if len(data) < 12:
            raise ValueError("passed raw parameter data size smaller than 12 bytes")
        try:
            Ptc(data[8])
        except TypeError:
            raise ValueError(f"ptc with unknown raw value {data[8]}")
        return cls(
            fsfw_param_id=FsfwParamId.unpack(data),
            param_raw=data[12:],
        )

    def parse_scalar_param(self) -> Union[int, float]:
        return parse_scalar_param(self)


def parse_scalar_param(wrapper: Parameter) -> Union[int, float]:
    return deserialize_scalar_entry(wrapper.ptc, wrapper.pfc, wrapper.param_raw)


__BASE_LEN_ERR = "Invalid parameter data size, smaller than "


def __deserialize_unsigned_scalar_entry(
    pfc: int, ptc: int, param_data: bytes, param_len: int
) -> int:
    if pfc == PfcUnsigned.ONE_BYTE:
        if param_len < 1:
            raise ValueError(f"{__BASE_LEN_ERR} 1")
        return param_data[0]
    elif pfc == PfcUnsigned.TWO_BYTES:
        if param_len < 2:
            raise ValueError(f"{__BASE_LEN_ERR} 2")
        return struct.unpack("!H", param_data[0:2])[0]
    if pfc == PfcUnsigned.FOUR_BYTES:
        if param_len < 4:
            raise ValueError(f"{__BASE_LEN_ERR} 4")
        return struct.unpack("!I", param_data[0:4])[0]
    elif pfc == PfcUnsigned.EIGHT_BYTES:
        if param_len < 8:
            raise ValueError(f"{__BASE_LEN_ERR} 8")
        return struct.unpack("!Q", param_data[0:8])[0]
    else:
        raise NotImplementedError(f"Parsing of unsigned PTC {ptc} not implemented for PFC {pfc}")


def __deserialize_signed_scalar_entry(pfc: int, ptc: int, tm_data: bytes, param_len: int) -> int:
    if pfc == PfcSigned.ONE_BYTE:
        if param_len < 1:
            raise ValueError(f"{__BASE_LEN_ERR} 1")
        return struct.unpack("!b", tm_data[0:1])[0]
    elif pfc == PfcSigned.TWO_BYTES:
        if param_len < 2:
            raise ValueError(f"{__BASE_LEN_ERR} 2")
        return struct.unpack("!h", tm_data[0:2])[0]
    elif pfc == PfcSigned.FOUR_BYTES:
        if param_len < 4:
            raise ValueError(f"{__BASE_LEN_ERR} 4")
        return struct.unpack("!i", tm_data[0:4])[0]
    elif pfc == PfcSigned.EIGHT_BYTES:
        if param_len < 8:
            raise ValueError(f"{__BASE_LEN_ERR} 8")
        return struct.unpack("!q", tm_data[0:8])[0]
    else:
        raise NotImplementedError(f"Parsing of signed PTC {ptc} not implemented for PFC {pfc}")


def deserialize_scalar_entry(ptc: int, pfc: int, param_data: bytes) -> Union[int, float]:
    """Try to deserialize a scalar parameter entry (row = 1 and column = 1).

    :raises ValueError: Passed parameter data length invalid.
    :raises NotImplementedError: Parsing of parameter data not implemented for given PTC and PFC.
    """
    param_len = len(param_data)
    if param_len == 0:
        raise ValueError("passed parameter data is empty")
    if ptc == Ptc.UNSIGNED:
        return __deserialize_unsigned_scalar_entry(pfc, ptc, param_data, param_len)
    elif ptc == Ptc.SIGNED:
        return __deserialize_signed_scalar_entry(pfc, ptc, param_data, param_len)
    if ptc == Ptc.REAL:
        if pfc == PfcReal.FLOAT_SIMPLE_PRECISION_IEEE:
            if param_len < 4:
                raise ValueError(f"{__BASE_LEN_ERR} 4")
            return struct.unpack("!f", param_data[0:4])[0]
        elif pfc == PfcReal.DOUBLE_PRECISION_IEEE:
            if param_len < 8:
                raise ValueError(f"{__BASE_LEN_ERR} 8")
            return struct.unpack("!d", param_data[0:8])[0]
    raise NotImplementedError(
        f"Parsing of real (floating point) PTC {ptc} not implemented " f"for PFC {pfc}"
    )


def create_scalar_boolean_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameter: bool
) -> Parameter:
    return create_scalar_u8_parameter(object_id, domain_id, unique_id, int(parameter))


def create_scalar_u8_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameter: int
) -> Parameter:
    if parameter < 0 or parameter > pow(2, 8) - 1:
        raise ValueError(f"parameter {parameter} is not a valid u8")
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.UNSIGNED,
            pfc=PfcUnsigned.ONE_BYTE,
            rows=1,
            columns=1,
        ),
        param_raw=bytes([parameter]),
    )


def create_scalar_i8_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameter: int
) -> Parameter:
    if abs(parameter) > pow(2, 7) - 1:
        raise ValueError(f"parameter {parameter} is not a valid i8")
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.SIGNED,
            pfc=PfcSigned.ONE_BYTE,
            rows=1,
            columns=1,
        ),
        param_raw=struct.pack("b", parameter),
    )


def create_scalar_u16_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameter: int
) -> Parameter:
    if parameter < 0 or parameter > pow(2, 16) - 1:
        raise ValueError(f"parameter {parameter} is not a valid u16")
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.UNSIGNED,
            pfc=PfcUnsigned.TWO_BYTES,
            rows=1,
            columns=1,
        ),
        param_raw=struct.pack("!H", parameter),
    )


def create_scalar_i16_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameter: int
) -> Parameter:
    if abs(parameter) > pow(2, 15) - 1:
        raise ValueError(f"parameter {parameter} is not a valid i16")
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.SIGNED,
            pfc=PfcSigned.TWO_BYTES,
            rows=1,
            columns=1,
        ),
        param_raw=struct.pack("!h", parameter),
    )


def create_scalar_u32_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameter: int
) -> Parameter:
    if parameter < 0 or parameter > pow(2, 32) - 1:
        raise ValueError(f"parameter {parameter} is not a valid u32")
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.UNSIGNED,
            pfc=PfcUnsigned.FOUR_BYTES,
            rows=1,
            columns=1,
        ),
        param_raw=struct.pack("!I", parameter),
    )


def create_scalar_i32_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameter: int
) -> Parameter:
    if abs(parameter) > pow(2, 31) - 1:
        raise ValueError(f"parameter {parameter} is not a valid i32")
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.SIGNED,
            pfc=PfcSigned.FOUR_BYTES,
            rows=1,
            columns=1,
        ),
        param_raw=struct.pack("!i", parameter),
    )


def create_scalar_double_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameter: float
) -> Parameter:
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.REAL,
            pfc=PfcReal.DOUBLE_PRECISION_IEEE,
            rows=1,
            columns=1,
        ),
        param_raw=struct.pack("!d", parameter),
    )


def create_scalar_float_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameter: float
) -> Parameter:
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.REAL,
            pfc=PfcReal.FLOAT_SIMPLE_PRECISION_IEEE,
            rows=1,
            columns=1,
        ),
        param_raw=struct.pack("!f", parameter),
    )


def create_vector_float_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameters: Sequence[float]
):
    if not parameters:
        raise ValueError("passed parameter vector is empty or invalid")
    param_raw = bytearray()
    for param in parameters:
        param_raw.extend(struct.pack("!f", param))
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.REAL,
            pfc=PfcReal.FLOAT_SIMPLE_PRECISION_IEEE,
            rows=1,
            columns=len(parameters),
        ),
        param_raw=param_raw,
    )


def create_vector_double_parameter(
    object_id: bytes, domain_id: int, unique_id: int, parameters: Sequence[float]
) -> Parameter:
    if not parameters:
        raise ValueError("passed parameter vector is empty or invalid")
    param_raw = bytearray()
    for param in parameters:
        param_raw.extend(struct.pack("!d", param))
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.REAL,
            pfc=PfcReal.DOUBLE_PRECISION_IEEE,
            rows=1,
            columns=len(parameters),
        ),
        param_raw=param_raw,
    )


def create_matrix_float_parameter(
    object_id: bytes,
    domain_id: int,
    unique_id: int,
    parameters: Sequence[Sequence[float]],
) -> Parameter:
    if not parameters:
        raise ValueError("passed parameter matrix is empty or invalid")
    rows = len(parameters)
    columns = len(parameters[0])
    param_raw = bytearray()
    for param_row in parameters:
        if len(param_row) != columns:
            raise ValueError("rows in passed mastrix do not have uniform number of columns")
        for val_at_column in param_row:
            param_raw.extend(struct.pack("!f", val_at_column))
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.REAL,
            pfc=PfcReal.FLOAT_SIMPLE_PRECISION_IEEE,
            rows=rows,
            columns=columns,
        ),
        param_raw=param_raw,
    )


def create_matrix_double_parameter(
    object_id: bytes,
    domain_id: int,
    unique_id: int,
    parameters: Sequence[Sequence[float]],
) -> Parameter:
    if not parameters:
        raise ValueError("passed parameter matrix is empty or invalid")
    rows = len(parameters)
    columns = len(parameters[0])
    param_raw = bytearray()
    for param_row in parameters:
        if len(param_row) != columns:
            raise ValueError("rows in passed mastrix do not have uniform number of columns")
        for val_at_column in param_row:
            param_raw.extend(struct.pack("!d", val_at_column))
    return Parameter(
        fsfw_param_id=FsfwParamId(
            object_id=object_id,
            param_id=ParameterId(domain_id, unique_id, 0),
            ptc=Ptc.REAL,
            pfc=PfcReal.DOUBLE_PRECISION_IEEE,
            rows=rows,
            columns=columns,
        ),
        param_raw=param_raw,
    )



from __future__ import annotations
import enum
from typing import Any, cast, Type, Optional

from tmtccmd.cfdp import CfdpRequestWrapper


class TcProcedureType(enum.Enum):
    TREE_COMMANDING = 0
    CFDP = 1
    CUSTOM = 2


class TcProcedureBase:
    def __init__(self, ptype: TcProcedureType):
        self.ptype = ptype


class CustomProcedureInfo(TcProcedureBase):
    def __init__(self, procedure: Any):
        super().__init__(TcProcedureType.CUSTOM)
        self.procedure = procedure

    def __repr__(self):
        return f"{self.__class__.__name__}(info={self.procedure!r}"


class TreeCommandingProcedure(TcProcedureBase):
    """Generic abstraction for procedures. A procedure can be a single command or a sequence
    of commands. Generally, one procedure is mapped to a specific TC queue which is packed
    during run-time"""

    def __init__(self, cmd_path: Optional[str]):
        super().__init__(TcProcedureType.TREE_COMMANDING)
        self.cmd_path = cmd_path

    @classmethod
    def empty(cls):
        return cls(None)

    def __repr__(self):
        return f"CmdInfo(cmd_path={self.cmd_path!r})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TreeCommandingProcedure):
            return self.cmd_path == other.cmd_path
        return False


class CfdpProcedure(TcProcedureBase):
    def __init__(self):
        super().__init__(TcProcedureType.CFDP)
        self.request_wrapper = CfdpRequestWrapper(None)

    @property
    def cfdp_request_type(self):
        return self.request_wrapper.request


class ProcedureWrapper:
    """Procedure helper class. It wraps the concrete procedure object but allows easily casting
    it to concrete types supported by the framework."""

    def __init__(self, procedure: Optional[TcProcedureBase]):
        self.procedure = procedure

    def __repr__(self):
        return f"{self.__class__.__name__}(base={self.procedure!r})"

    @property
    def proc_type(self):
        assert self.procedure is not None
        return self.procedure.ptype

    def __cast_internally(
        self,
        obj_type: Type[TcProcedureBase],
        obj: TcProcedureBase,
        expected_type: TcProcedureType,
    ) -> Any:
        assert self.procedure is not None
        if obj.ptype != expected_type:
            raise TypeError(f"Invalid object {obj} for type {self.procedure.ptype}")
        return cast(obj_type, obj)

    def to_tree_commanding_procedure(self) -> TreeCommandingProcedure:
        assert self.procedure is not None
        return self.__cast_internally(
            TreeCommandingProcedure, self.procedure, TcProcedureType.TREE_COMMANDING
        )

    def to_cfdp_procedure(self) -> CfdpProcedure:
        assert self.procedure is not None
        return self.__cast_internally(CfdpProcedure, self.procedure, TcProcedureType.CFDP)

    def to_custom_procedure(self) -> CustomProcedureInfo:
        assert self.procedure is not None
        return self.__cast_internally(CustomProcedureInfo, self.procedure, TcProcedureType.CUSTOM)



from tmtccmd.util.obj_id import ComponentIdMapping, ObjectIdU32


INVALID_ID = bytes([0xFF, 0xFF, 0xFF, 0xFF])


def get_base_component_id_mapping() -> ComponentIdMapping:
    """These are the object IDs for the tmtccmd core. The core will usually take care of
    inserting these into the object manager during the program initialization.

    :return Dictionary of the core object IDs
    """
    invalid_id = ObjectIdU32.from_bytes(raw=INVALID_ID)
    invalid_id.name = "Invalid ID"
    object_id_dict = {INVALID_ID: invalid_id}
    return object_id_dict


def get_core_object_ids() -> ComponentIdMapping:
    return get_base_component_id_mapping()



import json
import logging

from tmtccmd.util.conf_util import wrapped_prompt
from tmtccmd.util.json import check_json_file, JsonKeyNames
from tmtccmd.config.defs import ComIfDictT, CoreComInterfaces


def determine_com_if(com_if_dict: ComIfDictT, json_cfg_path: str, use_prompts: bool) -> str:
    do_prompt_com_if = False
    com_if_string = ""
    if not check_json_file(json_cfg_path):
        do_prompt_com_if = True
    if not do_prompt_com_if:
        with open(json_cfg_path, "r") as read:
            com_if_string = ""
            try:
                load_data = json.load(read)
                com_if_string = load_data[JsonKeyNames.COM_IF.value]
            except KeyError:
                do_prompt_com_if = True
            com_if_string = str(com_if_string)
    if do_prompt_com_if and use_prompts:
        com_if_string = prompt_com_if(com_if_dict)
        save_to_json = wrapped_prompt("Do you want to store the communication interface? ([Y]/n): ")
        if save_to_json.lower() in ["", "y", "yes", "1"]:
            store_com_if_json(com_if_string=com_if_string, json_cfg_path=json_cfg_path)
    elif do_prompt_com_if and not use_prompts:
        return CoreComInterfaces.DUMMY.value
    return com_if_string


def prompt_com_if(com_if_dict: ComIfDictT) -> str:
    com_if_string = ""
    while True:
        com_if_list = []
        print("List of available communication interfaces:")
        for index, (name, com_if_value) in enumerate(com_if_dict.items()):
            print(f"{index}: {com_if_value[0]}")
            com_if_list.append(name)
        com_if_key = wrapped_prompt("Please enter the desired communication interface by key: ")
        if not com_if_key.isdigit():
            print("Key is not a digit, try again")
            continue
        com_if_key = int(com_if_key)
        if com_if_key >= len(com_if_list):
            print("Key invalid, try again.")
            continue
        com_if_string = com_if_list[com_if_key]
        break
    return com_if_string


def store_com_if_json(com_if_string: str, json_cfg_path: str):
    with open(json_cfg_path, "r+") as file:
        data = json.load(file)
        data[JsonKeyNames.COM_IF.value] = com_if_string
        file.seek(0)
        json.dump(data, file, indent=4)
    logger = logging.getLogger(__name__)
    logger.info(f"Communication interface was stored in the JSON file {json_cfg_path}")
    logger.info("Delete this file or edit it manually to edit the communication interface")



"""Contains definitions and functions related to PUS Service 20 Telecommands."""

from __future__ import annotations

import struct
from typing import Optional

from deprecated.sphinx import deprecated
from spacepackets.ecss import PusService
from spacepackets.ecss.fields import PfcReal, PfcUnsigned, Ptc
from spacepackets.ecss.tc import PusTelecommand

from tmtccmd.pus.s20_fsfw_param_defs import (  # noqa: F401
    CustomSubservice,
    FsfwParamId,
    Parameter,
    create_matrix_double_parameter,
    create_matrix_float_parameter,
    create_scalar_boolean_parameter,
    create_scalar_double_parameter,
    create_scalar_float_parameter,
    create_scalar_i8_parameter,
    create_scalar_i16_parameter,
    create_scalar_i32_parameter,
    create_scalar_u8_parameter,
    create_scalar_u16_parameter,
    create_scalar_u32_parameter,
    create_vector_double_parameter,
    create_vector_float_parameter,
)


def create_load_param_cmd(parameter: Parameter, apid: int = 0) -> PusTelecommand:
    return PusTelecommand(
        apid=apid,
        service=PusService.S20_PARAMETER,
        subservice=CustomSubservice.TC_LOAD,
        app_data=parameter.pack(),
    )


def create_dump_param_cmd(param_fsfw_id: FsfwParamId, apid: int = 0) -> PusTelecommand:
    return PusTelecommand(
        apid=apid,
        service=PusService.S20_PARAMETER,
        subservice=CustomSubservice.TC_LOAD,
        app_data=param_fsfw_id.pack(),
    )


def create_load_param_cmd_from_raw(parameter_raw: bytes, apid: int = 0) -> PusTelecommand:
    return PusTelecommand(
        apid=apid,
        service=PusService.S20_PARAMETER,
        subservice=CustomSubservice.TC_LOAD,
        app_data=parameter_raw,
    )


@deprecated(
    version="3.1.0",
    reason="use create_scalar_boolean_parameter instead",
)
def pack_boolean_parameter_app_data(
    object_id: bytes, domain_id: int, unique_id: int, parameter: bool
) -> Optional[bytearray]:
    return pack_scalar_boolean_parameter_app_data(object_id, domain_id, unique_id, parameter)


@deprecated(
    version="4.0.0a2",
    reason="Please use create_scalar_boolean_parameter instead",
)
def pack_scalar_boolean_parameter_app_data(
    object_id: bytes, domain_id: int, unique_id: int, parameter: bool
) -> Optional[bytearray]:
    """Tailored towards FSFW applications.

    :param object_id:
    :param domain_id:
    :param unique_id:
    :param parameter:
    :return: Application data
    """
    return pack_scalar_u8_parameter_app_data(object_id, domain_id, unique_id, int(parameter))


@deprecated(
    version="4.0.0a2",
    reason="Please use create_scalar_u8_parameter instead",
)
def pack_scalar_u8_parameter_app_data(
    object_id: bytes, domain_id: int, unique_id: int, parameter: int
) -> Optional[bytearray]:
    """Tailored towards FSFW applications.

    :param object_id:
    :param domain_id:
    :param unique_id:
    :param parameter:
    :return: Application data
    """
    data_to_pack = prepare_param_packet_header(
        object_id=object_id,
        domain_id=domain_id,
        unique_id=unique_id,
        ptc=Ptc.UNSIGNED,
        pfc=PfcUnsigned.ONE_BYTE,
        rows=1,
        columns=1,
    )
    if data_to_pack is not None:
        data_to_pack.append(parameter)
    return data_to_pack


@deprecated(
    version="4.0.0a2",
    reason="Please use create_scalar_double_parameter instead",
)
def pack_scalar_double_param_app_data(
    object_id: bytes, domain_id: int, unique_id: int, parameter: float
) -> Optional[bytearray]:
    data_to_pack = prepare_param_packet_header(
        object_id=object_id,
        domain_id=domain_id,
        unique_id=unique_id,
        ptc=Ptc.REAL,
        pfc=PfcReal.DOUBLE_PRECISION_IEEE,
        rows=1,
        columns=1,
    )
    if data_to_pack is not None:
        data_to_pack.extend(struct.pack("!d", parameter))
    return data_to_pack


@deprecated(
    version="4.0.0a2",
    reason="use create_scalar_float_parameter instead",
)
def pack_scalar_float_param_app_data(
    object_id: bytes, domain_id: int, unique_id: int, parameter: float
) -> Optional[bytearray]:
    data_to_pack = prepare_param_packet_header(
        object_id=object_id,
        domain_id=domain_id,
        unique_id=unique_id,
        ptc=Ptc.REAL,
        pfc=PfcReal.FLOAT_SIMPLE_PRECISION_IEEE,
        rows=1,
        columns=1,
    )
    if data_to_pack is not None:
        data_to_pack.extend(struct.pack("!f", parameter))
    return data_to_pack


@deprecated(
    version="4.0.0a2",
    reason="use ParamWrapper helper class with pack() instead",
)
def prepare_param_packet_header(
    object_id: bytes,
    domain_id: int,
    unique_id: int,
    ptc: Ptc,
    pfc: int,
    rows: int,
    columns: int,
    start_at_idx: int = 0,
) -> Optional[bytearray]:
    parameter_id = bytearray(4)
    parameter_id[0] = domain_id
    if unique_id > 255:
        raise ValueError("Invalid unique ID, should be smaller than 255!")
    parameter_id[1] = unique_id
    parameter_id[2] = (start_at_idx >> 8) & 0xFF
    parameter_id[3] = start_at_idx & 0xFF
    data_to_pack = bytearray(object_id)
    data_to_pack.extend(parameter_id)
    data_to_pack.extend(pack_type_and_matrix_data(ptc=ptc, pfc=pfc, rows=rows, columns=columns))
    return data_to_pack


@deprecated(
    version="4.0.0a2",
    reason="use ParamWrapper helper class with pack() instead",
)
def pack_type_and_matrix_data(ptc: int, pfc: int, rows: int, columns: int) -> bytearray:
    data = bytearray(4)
    data[0] = ptc
    data[1] = pfc
    data[2] = rows
    data[3] = columns
    return data


@deprecated(
    version="4.0.0a2",
    reason="use ParameterId helper class with pack() instead",
)
def pack_parameter_id(domain_id: int, unique_id: int, linear_index: int) -> bytearray:
    """Packs the Parameter ID (bytearray with 4 bytes) which is part of the service 20 packets.
    The first byte of the parameter ID is the domain ID, the second byte is a unique ID and the
    last two bytes are a linear index if a parameter is not loaded from index 0.
    :param domain_id:       One byte domain ID
    :param unique_id:       One byte unique ID
    :param linear_index:    Two byte linear index.
    """
    parameter_id = bytearray(4)
    parameter_id[0] = domain_id
    parameter_id[1] = unique_id
    parameter_id[2] = linear_index >> 8 & 0xFF
    parameter_id[3] = linear_index & 0xFF
    return parameter_id



"""Contains classes and functions that perform all printing functionalities."""

import logging
import enum
from collections.abc import Generator
from typing import List, Optional

from spacepackets.util import get_printable_data_string, PrintFormats

from tmtccmd.tmtc.tm_base import PusTmInfoInterface, PusTmInterface
from tmtccmd.util.obj_id import ObjectIdU32
from tmtccmd.pus.tm.s3_hk_base import HkContentType
from tmtccmd.logging import get_current_time_string

_LOGGER = logging.getLogger(__name__)


class DisplayMode(enum.Enum):
    """List of display modes"""

    SHORT = enum.auto
    LONG = enum.auto


def get_validity_buffer_str(validity_buffer: bytes, num_vars: int) -> str:
    """
    :param validity_buffer: Validity buffer in bytes format
    :param num_vars: Number of variables
    :return:
    """
    valid_list = []
    counter = 0
    for _index, byte in enumerate(validity_buffer):
        for bit in range(1, 9):
            if FsfwTmTcPrinter.bit_extractor(byte, bit) == 1:
                valid_list.append(True)
            else:
                valid_list.append(False)
            counter += 1
            if counter == num_vars:
                break
    validity_lists = list(FsfwTmTcPrinter.chunks(n=16, lst=valid_list))
    for valid_list in validity_lists:
        printout = "Valid: ["
        for idx, valid in enumerate(valid_list):
            if valid:
                printout += "Y"
            else:
                printout += "N"
            if idx < len(valid_list) - 1:
                printout += ","
            else:
                printout += "]"
        return printout
    return ""


class FsfwTmTcPrinter:
    """This class handles printing to the command line and to files"""

    def __init__(
        self,
        file_logger: Optional[logging.Logger],
        display_mode: DisplayMode = DisplayMode.LONG,
    ):
        """
        :param display_mode:
        """
        self.display_mode = display_mode
        self.file_logger = file_logger

    @staticmethod
    def generic_short_string(packet_if: PusTmInterface) -> str:
        return f"Got TM[{packet_if.service}, {packet_if.subservice}]"

    def handle_long_tm_print(self, packet_if: PusTmInterface, info_if: PusTmInfoInterface):
        """Main function to print the most important information inside the telemetry
        :param packet_if: Core packet interface
        :param info_if: Information interface
        :return:
        """
        base_string = "Got TM: " + info_if.get_print_info()
        _LOGGER.info(base_string)
        if self.file_logger is not None:
            self.file_logger.info(f"{get_current_time_string(True)}: {base_string}")
        try:
            self.__handle_column_header_print(info_if=info_if)
            self.__handle_tm_content_print(info_if=info_if)
            self.__handle_additional_printout(info_if=info_if)
        except TypeError:
            _LOGGER.exception(
                "Type Error when trying to print TM Packet "
                f"[{packet_if.service}, {packet_if.subservice}]"
            )

    def __handle_column_header_print(self, info_if: PusTmInfoInterface):
        header_list = []
        info_if.append_telemetry_column_headers(header_list=header_list)
        print(header_list)
        if self.file_logger is not None:
            self.file_logger.info(header_list)

    def __handle_tm_content_print(self, info_if: PusTmInfoInterface):
        """
        :param info_if: Information interface
        :return:
        """
        content_list = []
        info_if.append_telemetry_content(content_list=content_list)
        print(content_list)
        if self.file_logger is not None:
            self.file_logger.info(content_list)

    def __handle_additional_printout(self, info_if: PusTmInfoInterface):
        additional_printout = info_if.get_custom_printout()
        if additional_printout is not None and additional_printout != "":
            _LOGGER.info(additional_printout)
            if self.file_logger is not None:
                print(additional_printout)

    def generic_hk_tm_print(
        self,
        content_type: HkContentType,
        object_id: ObjectIdU32,
        set_id: int,
        hk_data: bytes,
    ):
        """This function pretty prints HK packets with a given header and content list
        :param content_type: Type of content for HK packet
        :param object_id: Object ID of the HK source
        :param set_id: Unique set ID for the HK packet
        :param hk_data: User defined HK data
        :return:
        """
        if content_type == HkContentType.HK:
            print_prefix = "Housekeeping data"
        elif content_type == HkContentType.DEFINITIONS:
            print_prefix = "Housekeeping definitions"
        else:
            print_prefix = "Unknown housekeeping data"
        if object_id.name == "":
            object_id.name = "Unknown Name"
        generic_info = (
            f"{print_prefix} from Object ID"
            f" {object_id.name} ({object_id.as_hex_string}) with Set ID {set_id} and"
            f" {len(hk_data)} bytes of HK data"
        )
        _LOGGER.info(generic_info)
        if self.file_logger is not None:
            self.file_logger.info(f"{get_current_time_string(True)}: {generic_info}")

    def print_validity_buffer(self, validity_buffer: bytes, num_vars: int):
        printout = FsfwTmTcPrinter.get_validity_buffer_str(validity_buffer, num_vars)
        print(printout)
        if self.file_logger:
            self.file_logger.info(printout)

    @staticmethod
    def get_validity_buffer_str(validity_buffer: bytes, num_vars: int) -> str:
        """
        :param validity_buffer: Validity buffer in bytes format
        :param num_vars: Number of variables
        :return:
        """
        return get_validity_buffer_str(validity_buffer, num_vars)

    @staticmethod
    def bit_extractor(byte: int, position: int):
        """

        :param byte:
        :param position:
        :return:
        """
        shift_number = position + (6 - 2 * (position - 1))
        return (byte >> shift_number) & 1

    @staticmethod
    def print_data(data: bytes):
        """
        :param data: Data to print
        :return: None
        """
        string = get_printable_data_string(print_format=PrintFormats.HEX, data=data)
        _LOGGER.info(string)

    @staticmethod
    def chunks(lst: List, n) -> Generator[List[List], None, None]:
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i : i + n]



from abc import abstractmethod, ABC

from tmtccmd.com import ComInterface
from tmtccmd.tmtc.procedure import ProcedureWrapper
from tmtccmd.tmtc.queue import QueueWrapper, QueueEntryHelper


class FeedWrapper:
    """This class wraps the queue and some additional information and useful fields which
    can be set by the user.

    :var queue_helper: Can be used to simplify insertion of queue entries like telecommands into
        the queue
    :var dispatch_next_queue: Can be used to prevent the dispatch of the queue
    :var modes: Currently contains the current TC Mode and TM mode of the calling handler class
    """

    def __init__(self, queue_wrapper: QueueWrapper, auto_dispatch: bool):
        from tmtccmd.core.base import ModeWrapper

        self.queue_wrapper = queue_wrapper
        self.dispatch_next_queue = auto_dispatch
        self.modes = ModeWrapper()


class SendCbParams:
    """Wrapper for all important parameters passed to the TC send callback.

    :var info: Procedure info about the procedure this queue entry is related too
    :var entry: Queue entry base type. The user can cast this back to the concrete
            type or just use duck typing if the concrete type is known
    :var com_if: Communication interface. Will generally be used to send the packet,
            using the :py:func:`tmtccmd.com_if.ComInterface.send` method
    """

    def __init__(self, info: ProcedureWrapper, entry: QueueEntryHelper, com_if: ComInterface):
        """Creates the parameters passed to the send callback."""
        self.info = info
        self.entry = entry
        self.com_if = com_if


class TcHandlerBase(ABC):
    """Generic abstract class for a TC handler object. Should be implemented by the user.
    This object then takes care of sending packets by providing the :py:meth:`send_cb`
    send-callback. It also provides telecommand queues by providing the :py:meth:`feed_cb` queue
    feeder callback.
    """

    def __init__(self):
        pass

    @abstractmethod
    def send_cb(self, send_params: SendCbParams):
        """This function callback will be called for each queue entry. This also includes
        miscellaneous queue entries, for example the ones used to log additional information.
        It is up to the user code implementation to determine the concrete queue entry and what
        to do with it.

        In general, an implementation will perform the following steps:

        1. Determine the queue entry and what to do with it
        2. If applicable, retrieve the raw data to send from the queue entry and send it using
           the generic communication interface

        All delay related entries will generally be handled by the send queue consumer so there
        is no need to manually delay the application in this callback. However, the queue consumer
        will not handle log entries so the user needs to take care of handling these
        entries and log the content to a console, file logger or any other system used to log
        something.

        :param send_params:
        """
        pass

    @abstractmethod
    def queue_finished_cb(self, info: ProcedureWrapper):
        pass

    @abstractmethod
    def feed_cb(self, info: ProcedureWrapper, wrapper: FeedWrapper):
        """This function will be called to retrieve a telecommand queue from the user code, based
        on a procedure. The passed feed wrapper can be used to set the TC queue or other
        parameter like the inter-packet delay.

        :param info: Generic base class for a procedure. For example, the
            py:class:`tmtccmd.tmtc.DefaultProcedureInfo` class uses a service string
            and op code string which can be used in the user code to select between different
            telecommand queues being packed
        :param wrapper: Wrapper type around the queue. It also contains a queue helper class
            to simplify adding entries to the telecommand queue
        :return:
        """
        pass



import dataclasses
import functools

from tmtccmd.tmtc import TcHandlerBase
from tmtccmd.tmtc.procedure import TreeCommandingProcedure
from tmtccmd.tmtc.queue import DefaultPusQueueHelper

SERVICE_HANDLER_DICT = dict()


@dataclasses.dataclass
class ServiceProviderParams:
    handler_base: TcHandlerBase
    info: TreeCommandingProcedure
    queue_helper: DefaultPusQueueHelper
    op_code: str


def service_provider(service: str):
    """Decorator. TODO: Documentation"""
    global SERVICE_HANDLER_DICT

    def actual_service_decorator(handler):
        global SERVICE_HANDLER_DICT

        @functools.wraps(handler)
        def service_handler_wrapper(p: ServiceProviderParams):
            return handler(p)

        SERVICE_HANDLER_DICT.update({service: service_handler_wrapper})
        return service_handler_wrapper

    return actual_service_decorator


def route_to_registered_service_handlers(service: str, p: ServiceProviderParams) -> bool:
    if service in SERVICE_HANDLER_DICT:
        SERVICE_HANDLER_DICT[service](p)
        return True
    return False



"""Used to send multiple TCs in sequence"""

import enum
import logging
from datetime import timedelta
from typing import Optional

from spacepackets.countdown import Countdown

from tmtccmd.com import ComInterface
from tmtccmd.tmtc import (
    ProcedureWrapper,
    QueueEntryHelper,
    TcQueueEntryBase,
    TcQueueEntryType,
)
from tmtccmd.tmtc.handler import SendCbParams, TcHandlerBase
from tmtccmd.tmtc.queue import QueueWrapper


class SenderMode(enum.IntEnum):
    BUSY = 0
    DONE = 1


class SeqResultWrapper:
    def __init__(self, mode: SenderMode):
        self.mode = mode
        self.longest_rem_delay: timedelta = timedelta()
        self.tc_sent: bool = False
        self.queue_empty: bool = False
        self.next_entry_is_tc: bool = False


class SequentialCcsdsSender:
    """Specific implementation of CommandSenderReceiver to send multiple telecommands in sequence"""

    def __init__(
        self,
        queue_wrapper: QueueWrapper,
        tc_handler: TcHandlerBase,
    ):
        """
        :param queue_wrapper: Wrapper object containing the queue and queue handling properties
        :param tc_handler:
        """
        self._tc_handler = tc_handler
        self._queue_wrapper = queue_wrapper
        self._proc_wrapper = ProcedureWrapper(None)
        self._mode = SenderMode.DONE
        self._wait_cd = Countdown(None)
        self._send_cd = Countdown(queue_wrapper.inter_cmd_delay)
        self._current_res = SeqResultWrapper(self._mode)
        self._current_res.longest_rem_delay = queue_wrapper.inter_cmd_delay
        self._op_divider = 0
        self._last_queue_entry: Optional[TcQueueEntryBase] = None
        self._last_tc: Optional[TcQueueEntryBase] = None

    @property
    def queue_wrapper(self):
        return self._queue_wrapper

    @queue_wrapper.setter
    def queue_wrapper(self, queue_wrapper: QueueWrapper):
        """This setter throws a ValueError if the sequential sender is busy with another queue"""
        if self._mode == SenderMode.BUSY:
            raise ValueError("Busy with other queue")
        self._mode = SenderMode.BUSY
        # There is no need to delay sending of the first entry, the send delay is inter-packet
        # only
        self._send_cd.timeout = timedelta()
        self._current_res.longest_rem_delay = queue_wrapper.inter_cmd_delay
        self._proc_wrapper.procedure = self._queue_wrapper.info
        self._queue_wrapper = queue_wrapper

    def handle_new_queue_forced(self, queue_wrapper: QueueWrapper):
        self._mode = SenderMode.DONE
        self.queue_wrapper = queue_wrapper
        self._proc_wrapper.procedure = self._queue_wrapper.info

    def resume(self):
        """Can be used to resume a finished sequential sender it the provided queue is
        not empty anymore"""
        if self._mode == SenderMode.DONE and self.queue_wrapper.queue:
            self._mode = SenderMode.BUSY

    def operation(self, com_if: ComInterface) -> SeqResultWrapper:
        """Primary function which should be called periodically to consume a TC queue.

        :param com_if: Communication interface used to send telecommands. Will be passed to the
            user send function
        """
        self._handle_current_tc_queue(com_if)
        self._current_res.mode = self._mode
        return self._current_res

    @property
    def mode(self):
        return self._mode

    def _handle_current_tc_queue(self, com_if: ComInterface):
        """Primary function which is called for sequential transfer.
        :return:
        """
        # Do not use continue anywhere in this while loop for now
        if not self.queue_wrapper.queue:
            self._current_res.queue_empty = True
            if self.no_delay_remaining():
                self._proc_wrapper.procedure = self._queue_wrapper.info
                # cache this for last wait time
                self._tc_handler.queue_finished_cb(self._proc_wrapper)
                self._mode = SenderMode.DONE
                return
        else:
            self._current_res.queue_empty = False
            self._check_next_telecommand(com_if)
        self._update_largest_delay()
        self.__print_rem_timeout(op_divider=self._op_divider)
        self._op_divider += 1

    def __print_rem_timeout(self, op_divider: int, divisor: int = 15):
        if not self.__wait_cd_timed_out() and op_divider % divisor == 0:
            rem_time = self._wait_cd.remaining_time()
            if self._wait_cd.remaining_time() > timedelta():
                print(f"{rem_time.total_seconds():.01f} seconds wait time remaining")

    def _check_next_telecommand(self, com_if: ComInterface):
        """Sends the next telecommand and returns whether an actual telecommand was sent"""
        next_queue_entry = self.queue_wrapper.queue[0]
        is_tc = self.handle_non_tc_entry(next_queue_entry)
        consume_queue_entry = True
        if is_tc:
            if self.no_delay_remaining():
                self._current_res.tc_sent = True
            else:
                self._current_res.tc_sent = False
                consume_queue_entry = False
        else:
            self._current_res.tc_sent = False
        if consume_queue_entry:
            self._tc_handler.send_cb(
                SendCbParams(self._proc_wrapper, QueueEntryHelper(next_queue_entry), com_if)
            )
            if is_tc:
                if self.queue_wrapper.inter_cmd_delay != self._send_cd.timeout:
                    self._send_cd.reset(self.queue_wrapper.inter_cmd_delay)
                else:
                    self._send_cd.reset()
            self.queue_wrapper.queue.popleft()
            if self.queue_wrapper.queue:
                self._current_res.next_entry_is_tc = self.queue_wrapper.queue[0].is_tc()
            else:
                self._current_res.next_entry_is_tc = False
        if not self.queue_wrapper.queue and self.no_delay_remaining():
            self._tc_handler.queue_finished_cb(ProcedureWrapper(self._queue_wrapper.info))
            self._mode = SenderMode.DONE

    def no_delay_remaining(self) -> bool:
        return self.__send_cd_timed_out() and self.__wait_cd_timed_out()

    def __send_cd_timed_out(self):
        """Internal wrapper API to allow easier testing"""
        return self._send_cd.timed_out()

    def __wait_cd_timed_out(self):
        """Internal wrapper API to allow easier testing"""
        return self._wait_cd.timed_out()

    def handle_non_tc_entry(self, queue_entry: TcQueueEntryBase) -> bool:
        """
        Checks whether the entry in the pus_tc queue is a telecommand.
        :param queue_entry: Generic queue entry
        :return: True if queue entry is telecommand, False if it is not
        """
        if not isinstance(queue_entry, TcQueueEntryBase):
            raise ValueError("Invalid queue entry detected")
        cast_wrapper = QueueEntryHelper(queue_entry)
        if queue_entry.etype == TcQueueEntryType.WAIT:
            wait_entry = cast_wrapper.to_wait_entry()
            logging.getLogger(__name__).info(
                "Waiting for" f" {wait_entry.wait_time.total_seconds() * 1000} milliseconds."
            )
            self._wait_cd.reset(new_timeout=wait_entry.wait_time)
        elif queue_entry.etype == TcQueueEntryType.PACKET_DELAY:
            timeout_entry = cast_wrapper.to_packet_delay_entry()
            self.queue_wrapper.inter_cmd_delay = timeout_entry.delay_time
            self._send_cd.reset(new_timeout=timeout_entry.delay_time)
        is_tc = queue_entry.is_tc()
        if is_tc:
            self._last_tc = queue_entry
        self._last_queue_entry = queue_entry
        return is_tc

    def _update_largest_delay(self):
        self._current_res.longest_rem_delay = max(
            self._wait_cd.remaining_time(), self._send_cd.remaining_time()
        )



import struct
from unittest import TestCase
from tmtccmd.pus.s5_fsfw_event import Service5Tm, Subservice, EventDefinition


class TestSrv5Tm(TestCase):
    def setUp(self):
        self.apid = 0x08
        self.obj_id = bytes([0x00, 0x01, 0x02, 0x03])
        self.event_def = EventDefinition(event_id=5, reporter_id=self.obj_id, param1=22, param2=942)
        self.srv5_tm = Service5Tm(
            apid=self.apid,
            subservice=Subservice.TM_INFO_EVENT,
            event=self.event_def,
            timestamp=bytes(),
        )

    def test_basic(self):
        self.assertEqual(self.srv5_tm.sp_header, self.srv5_tm.pus_tm.space_packet_header)
        self.assertEqual(self.srv5_tm.service, 5)
        self.assertEqual(self.srv5_tm.subservice, Subservice.TM_INFO_EVENT)
        self.assertEqual(len(self.srv5_tm.source_data), 14)

    def test_packed(self):
        raw_tm = self.srv5_tm.pack()
        self.assertEqual(raw_tm[7], 5)
        self.assertEqual(raw_tm[8], Subservice.TM_INFO_EVENT)
        self.assertEqual(struct.unpack("!H", raw_tm[13:15])[0], 5)
        self.assertEqual(raw_tm[15:19], self.obj_id)
        self.assertEqual(struct.unpack("!I", raw_tm[19:23])[0], 22)
        self.assertEqual(struct.unpack("!I", raw_tm[23:27])[0], 942)

    def test_unpack(self):
        raw_tm = self.srv5_tm.pack()
        unpacked = Service5Tm.unpack(raw_tm, 0)
        self.assertEqual(self.srv5_tm, unpacked)
        self.assertEqual(unpacked.event_definition, self.event_def)



# -*- coding: utf-8 -*-
"""Contains classes and functions to deserialize PUS Service 5 Telemetry"""

from __future__ import annotations

import dataclasses
import struct

from spacepackets import SpacePacketHeader
from spacepackets.ccsds.spacepacket import PacketId, PacketSeqCtrl
from spacepackets.ecss.defs import PusService
from spacepackets.ecss.pus_5_event import Subservice
from spacepackets.ecss.tm import CdsShortTimestamp, AbstractPusTm, PusTelemetry
from tmtccmd.pus.s5_fsfw_event_defs import Severity


@dataclasses.dataclass
class EventDefinition:
    event_id: int
    reporter_id: bytes
    param1: int
    param2: int

    def pack(self) -> bytes:
        raw = bytearray(struct.pack("!H", self.event_id))
        if len(self.reporter_id) < 4:
            raise ValueError("reporter ID must be at least 4 bytes wide")
        raw.extend(self.reporter_id)
        raw.extend(struct.pack("!I", self.param1))
        raw.extend(struct.pack("!I", self.param2))
        return raw

    @classmethod
    def empty(cls) -> EventDefinition:
        return cls(0, bytes([0, 0, 0, 0]), 0, 0)

    @classmethod
    def from_bytes(cls, data: bytes) -> EventDefinition:
        if len(data) < 14:
            raise ValueError("full FSFW event definition must be at least 14 bytes wide")
        event_id = struct.unpack("!H", data[0:2])[0]
        object_id = bytes(data[2:6])
        param1 = struct.unpack("!I", data[6:10])[0]
        param2 = struct.unpack("!I", data[10:14])[0]
        return cls(event_id, object_id, param1, param2)


class Service5Tm(AbstractPusTm):
    def __init__(
        self,
        apid: int,
        subservice: Subservice,
        event: EventDefinition,
        timestamp: bytes,
        ssc: int = 0,
        packet_version: int = 0b000,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        """Create a FSFW tailored Event Service 5 telemetry instance.
        Use the unpack function to create an instance from a raw bytestream instead.
        :raises ValueError: Invalid input arguments
        """
        self.pus_tm = PusTelemetry(
            service=PusService.S5_EVENT,
            subservice=subservice,
            timestamp=timestamp,
            seq_count=ssc,
            source_data=event.pack(),
            apid=apid,
            packet_version=packet_version,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )

    @property
    def sp_header(self) -> SpacePacketHeader:
        return self.pus_tm.space_packet_header

    @property
    def timestamp(self) -> bytes:
        return self.pus_tm.timestamp

    def pack(self) -> bytearray:
        return self.pus_tm.pack()

    @property
    def service(self) -> int:
        return self.pus_tm.service

    @property
    def subservice(self) -> int:
        return self.pus_tm.subservice

    @property
    def ccsds_version(self) -> int:
        return self.pus_tm.ccsds_version

    @property
    def packet_id(self) -> PacketId:
        return self.pus_tm.packet_id

    @property
    def packet_seq_control(self) -> PacketSeqCtrl:
        return self.pus_tm.packet_seq_control

    @property
    def source_data(self) -> bytes:
        return self.pus_tm.source_data

    @classmethod
    def __empty(cls) -> Service5Tm:
        return cls(
            apid=0x0,
            subservice=Subservice.TM_INFO_EVENT,
            event=EventDefinition.empty(),
            timestamp=CdsShortTimestamp.empty().pack(),
        )

    @classmethod
    def from_tm(cls, pus_tm: PusTelemetry) -> Service5Tm:
        instance = cls.__empty()
        instance.pus_tm = pus_tm
        return instance

    @classmethod
    def unpack(cls, data: bytes, timestamp_len: int) -> Service5Tm:
        instance = cls.__empty()
        instance.pus_tm = PusTelemetry.unpack(data=data, timestamp_len=timestamp_len)
        return instance

    @property
    def severity(self) -> Severity:
        if self.subservice == Subservice.TM_INFO_EVENT:
            return Severity.INFO
        elif self.subservice == Subservice.TM_LOW_SEVERITY_EVENT:
            return Severity.LOW
        elif self.subservice == Subservice.TM_MEDIUM_SEVERITY_EVENT:
            return Severity.MEDIUM
        elif self.subservice == Subservice.TM_HIGH_SEVERITY_EVENT:
            return Severity.HIGH
        raise ValueError(f"invalid severity for subservice {self.subservice}")

    @property
    def event_definition(self) -> EventDefinition:
        return EventDefinition.from_bytes(self.pus_tm.source_data)

    def __eq__(self, other: object):
        if not isinstance(other, Service5Tm):
            return False
        return self.pus_tm == other.pus_tm



import warnings

from abc import abstractmethod

from spacepackets.ecss.tm import PusTelemetry
from spacepackets.util import PrintFormats

warnings.warn("the tm_base module is deprecated", DeprecationWarning, stacklevel=2)


class PusTmInterface:
    @abstractmethod
    def pack(self) -> bytearray:
        raise NotImplementedError

    @property
    @abstractmethod
    def tm_data(self) -> bytes:
        raise NotImplementedError

    @property
    @abstractmethod
    def valid(self) -> bool:
        raise NotImplementedError

    @property
    @abstractmethod
    def ssc(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def apid(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def service(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def subservice(self) -> int:
        raise NotImplementedError


class PusTmInfoInterface:
    @abstractmethod
    def get_print_info(self) -> str:
        return ""

    @abstractmethod
    def get_custom_printout(self) -> str:
        return ""

    @abstractmethod
    def get_source_data_string(self) -> str:
        return ""

    @abstractmethod
    def set_packet_info(self, print_info: str):
        pass

    @abstractmethod
    def append_packet_info(self, info: str):
        pass

    @abstractmethod
    def append_telemetry_column_headers(self, header_list: list):
        pass

    @abstractmethod
    def append_telemetry_content(self, content_list: list):
        pass


class PusTmBase(PusTmInterface):
    def __init__(self, pus_tm: PusTelemetry):
        self.pus_tm = pus_tm

    def pack(self) -> bytearray:
        return self.pus_tm.pack()

    @property
    def tm_data(self) -> bytes:
        return self.pus_tm.tm_data

    @property
    def ssc(self) -> int:
        return self.pus_tm.seq_count

    @property
    def apid(self) -> int:
        return self.pus_tm.apid

    @property
    def service(self) -> int:
        return self.pus_tm.service

    @property
    def subservice(self) -> int:
        return self.pus_tm.subservice


class PusTmInfoBase(PusTmInfoInterface):
    def __init__(self, pus_tm: PusTelemetry):
        self.pus_tm = pus_tm
        self._custom_printout = ""
        self._print_info = ""

    def get_print_info(self) -> str:
        return self._print_info

    def get_custom_printout(self) -> str:
        return self._custom_printout

    def set_custom_printout(self, custom_string: str):
        self._custom_printout = custom_string

    def get_source_data_string(self, print_format: PrintFormats = PrintFormats.HEX) -> str:
        return self.pus_tm.get_source_data_string(print_format=print_format)

    def set_packet_info(self, print_info: str):
        self._print_info = print_info

    def append_packet_info(self, info: str):
        self._print_info += info

    def append_telemetry_content(self, content_list: list):
        """Default implementation adds the PUS header content to the list which can then be
        printed with a simple print() command. To add additional content, override this method.
        Any child class should call this function as well if header information is required.

        :param content_list: Header content will be appended to this list
        :return:
        """
        content_list.append(f"{self.pus_tm.service}")
        content_list.append(f"{self.pus_tm.subservice}")
        content_list.append(f"{self.pus_tm.pus_tm_sec_header.message_counter}")
        content_list.append(f"{self.pus_tm.pus_tm_sec_header.time_provider.as_unix_seconds()}")
        content_list.append(f"{self.pus_tm.pus_tm_sec_header.time_provider.as_time_string()}")
        content_list.append(f"0x{self.pus_tm.space_packet_header.apid:02x}")
        content_list.append(f"{self.pus_tm.space_packet_header.seq_count}")

    def append_telemetry_column_headers(self, header_list: list):
        """Default implementation adds the PUS header content header (confusing, I know)
        to the list which can then be printed with a simple print() command.
        To add additional headers, override this method. Any child class should
        call this function as well if header information is required.

        :param header_list: Header content will be appended to this list
        :return:
        """
        header_list.append("Service")
        header_list.append("Subservice")
        header_list.append("MSG Counter")
        header_list.append("Time (Unix Seconds)")
        header_list.append("Time")
        header_list.append("APID")
        header_list.append("SSC")
        header_list.append("Packet valid")



import dataclasses
import logging
import threading
from collections import deque
from typing import List, Optional

import serial
from dle_encoder import DleEncoder, STX_CHAR, ETX_CHAR, DleErrorCodes

from tmtccmd.com import ComInterface
from tmtccmd.com.serial_base import SerialComBase, SerialCfg, SerialCommunicationType


@dataclasses.dataclass
class DleCfg:
    dle_queue_len: Optional[int] = None
    dle_max_frame: Optional[int] = None
    encode_cr: bool = True


class SerialDleComIF(SerialComBase, ComInterface):
    """Serial communication interface which uses the
    `DLE protocol <https://pypi.org/project/dle-encoder/>`_ to encode and decode packets.

    This class will spin up a receiver thread on the :meth:`open` call to poll for DLE encoded
    packets. This means that the :meth:`close` call might block until the receiver thread has shut
    down.
    """

    def __init__(self, ser_cfg: SerialCfg, dle_cfg: Optional[DleCfg]):
        super().__init__(
            logging.getLogger(__name__),
            ser_cfg=ser_cfg,
            ser_com_type=SerialCommunicationType.DLE_ENCODING,
        )
        self.dle_cfg = dle_cfg
        self.__encoder = DleEncoder()
        self.__reception_thread = None
        self.__reception_buffer = deque()
        self.__polling_shutdown: Optional[threading.Event] = threading.Event()

    @property
    def id(self) -> str:
        return self.ser_cfg.com_if_id

    def initialize(self, args: any = None) -> any:
        if self.dle_cfg and self.dle_cfg.dle_queue_len:
            self.__reception_buffer = deque(maxlen=self.dle_cfg.dle_queue_len)

    def open(self, args: any = None) -> None:
        """Spins up a receiver thread to permanently check for new DLE encoded packets."""
        super().open_port()
        self.__polling_shutdown.clear()
        self.__reception_thread = threading.Thread(target=self.__poll_dle_packets, daemon=True)
        self.__reception_thread.start()

    def __poll_dle_packets(self):
        # Poll permanently, but it is possible to join this thread every 200 ms
        self.serial.timeout = 0.2
        data = bytearray()
        while True:
            byte = self.serial.read()
            if len(byte) == 1:
                if byte[0] == STX_CHAR:
                    data.append(byte[0])
                    self.serial.timeout = 0.1
                    if self.dle_cfg and self.dle_cfg.dle_max_frame:
                        bytes_rcvd = self.serial.read_until(
                            serial.to_bytes([ETX_CHAR]), self.dle_cfg.dle_max_frame
                        )
                    else:
                        bytes_rcvd = self.serial.read_until(serial.to_bytes([ETX_CHAR]))
                    self.serial.timeout = 0.2
                    if bytes_rcvd[len(bytes_rcvd) - 1] == ETX_CHAR:
                        data.extend(bytes_rcvd)
                        # deque is thread-safe for appends and pops from and to the opposite side
                        self.__reception_buffer.appendleft(data)
                data = bytearray()
            elif self.__polling_shutdown.is_set():
                break

    def is_open(self) -> bool:
        return super().is_port_open()

    def close(self, args: any = None) -> None:
        self.__polling_shutdown.set()
        self.__reception_thread.join(0.4)
        super().close_port()

    def send(self, data: bytes):
        encoded_data = self.__encoder.encode(source_packet=data, add_stx_etx=True)
        self.serial.write(encoded_data)

    def receive(self, parameters: any = 0) -> List[bytes]:
        packet_list = []
        while self.__reception_buffer:
            data = self.__reception_buffer.pop()
            dle_retval, decoded_packet, read_len = self.__encoder.decode(source_packet=data)
            if dle_retval == DleErrorCodes.OK:
                packet_list.append(decoded_packet)
            else:
                self.logger.warning("DLE decoder error!")
        return packet_list

    def data_available(self, timeout: float, parameters: any = 0) -> int:
        return SerialComBase.data_available_from_queue(timeout, self.__reception_buffer)



from datetime import timedelta

from .base import ModeWrapper, BackendRequest
from tmtccmd.tmtc.ccsds_seq_sender import SeqResultWrapper, SenderMode


class BackendState:
    def __init__(
        self,
        mode_wrapper: ModeWrapper = ModeWrapper(),
        req: BackendRequest = BackendRequest.NONE,
    ):
        self._mode_wrapper = mode_wrapper
        self._req = req
        self._recommended_delay = timedelta()
        self._sender_res = SeqResultWrapper(SenderMode.DONE)

    @property
    def next_delay(self):
        return self._recommended_delay

    @property
    def request(self):
        return self._req

    @property
    def sender_res(self):
        return self._sender_res

    @property
    def tc_mode(self):
        return self._mode_wrapper.tc_mode

    @property
    def tm_mode(self):
        return self._mode_wrapper.tm_mode

    @property
    def mode_wrapper(self):
        return self._mode_wrapper



import json
import logging
import os
import enum
from typing import Any

_LOGGER = logging.getLogger(__name__)


class JsonKeyNames(enum.Enum):
    COM_IF = "com_if"
    TCPIP_UDP_DEST_IP_ADDRESS = "tcpip_udp_ip_addr"
    TCPIP_UDP_DEST_PORT = "tcpip_udp_port"
    TCPIP_UDP_RECV_IP_ADDRESS = "tcpip_udp_recv_addr"
    TCPIP_UDP_RECV_PORT = "tcpip_udp_recv_port"
    TCPIP_UDP_RECV_MAX_SIZE = "tcpip_udp_recv_max_size"

    TCPIP_TCP_DEST_IP_ADDRESS = "tcpip_tcp_ip_addr"
    TCPIP_TCP_DEST_PORT = "tcpip_tcp_port"
    TCPIP_TCP_RECV_MAX_SIZE = "tcpip_tcp_recv_max_size"

    SERIAL_BAUDRATE = "serial_baudrate"
    SERIAL_PORT = "serial_port"
    SERIAL_HINT = "serial_hint"


def check_json_file(json_cfg_path: str) -> bool:
    """The check JSON file and return whether it was valid or not. A JSON file is invalid
    if it does not exist or the format ins invalid.
    :return: True if JSON file is valid, False if not and a new one was created at the
    specified path
    """
    if json_cfg_path == "":
        json_cfg_path = "tmtc_config.json"
    if not os.path.isfile(json_cfg_path):
        with open(json_cfg_path, "w") as file:
            load_data = dict()
            json.dump(load_data, file)
            print(f"Configuration JSON {json_cfg_path} did not exist, created a new one.")
            return False
    else:
        with open(json_cfg_path, "r+") as file:
            try:
                json.load(file)
            except json.decoder.JSONDecodeError:
                _LOGGER.warning("JSON decode error, file format might be invalid. Replacing JSON")
                file.flush()
                file.truncate(0)
                file.seek(0)
                void_data = dict()
                json.dump(void_data, file)
                return False
    return True


def save_to_json_with_prompt(
    key: str, value: Any, name: str, json_cfg_path: str, json_obj: Any
) -> bool:
    save_to_json = input(f"Do you want to store the {name} to the configuration file? (y/n): ")
    if save_to_json.lower() in ["y", "yes"]:
        json_obj[key] = value
        _LOGGER.info(f"The {name} was stored to the JSON file {json_cfg_path}")
        _LOGGER.info("Delete this file or edit it manually to change it")
        return True
    return False



from deprecated.sphinx import deprecated
from spacepackets.ecss import PusService, PusTelecommand
from spacepackets.ecss.pus_17_test import Subservice


@deprecated(
    version="4.0.0a2",
    reason="use create... API instead",
)
def pack_service_17_ping_command(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return create_service_17_ping_command(apid, seq_count)


def create_service_17_ping_command(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    """Generate a simple ping PUS telecommand packet"""
    return PusTelecommand(
        service=PusService.S17_TEST,
        subservice=Subservice.TC_PING,
        apid=apid,
        seq_count=seq_count,
    )



import json
import logging
from typing import TextIO, Tuple

import serial
import serial.tools.list_ports
from tmtccmd.util.json import (
    check_json_file,
    JsonKeyNames,
    save_to_json_with_prompt,
)


_LOGGER = logging.getLogger(__name__)


def determine_baud_rate(json_cfg_path: str) -> int:
    """Determine baud rate. Tries to read from JSON first. If the baud rate is not contained
    in the config JSON, prompt it from user instead with the option to store value in JSON file.

    :return: Determined baud rate
    """
    baud_rate = 0
    prompt_baud_rate = False
    if not check_json_file(json_cfg_path=json_cfg_path):
        prompt_baud_rate = True

    if not prompt_baud_rate:
        with open(json_cfg_path, "r") as read:
            try:
                load_data = json.load(read)
                baud_rate = load_data[JsonKeyNames.SERIAL_BAUDRATE.value]
            except KeyError:
                prompt_baud_rate = True

    if prompt_baud_rate:
        while True:
            baud_rate = input("Please enter the baudrate for the serial protocol: ")
            if baud_rate.isdigit():
                baud_rate = int(baud_rate)
                break
            else:
                print("Invalid baud rate specified, try again.")
        with open(json_cfg_path, "r+") as json_file:
            json_obj = json.load(fp=json_file)
            if save_to_json_with_prompt(
                key=JsonKeyNames.SERIAL_BAUDRATE.value,
                value=baud_rate,
                json_cfg_path=json_cfg_path,
                json_obj=json_obj,
                name="baudrate",
            ):
                json_file.seek(0)
                json.dump(json_obj, json_file, indent=4)
    return baud_rate


def determine_com_port(json_cfg_path: str) -> str:
    """Determine serial port. Tries to read from JSON first. If the serial port is not contained
    in the config JSON, prompt it from user instead with the option to store value in JSON file.

    :return: Determined serial port
    """
    reconfig_com_port = False
    if not check_json_file(json_cfg_path=json_cfg_path):
        reconfig_com_port = True
    with open(json_cfg_path, "r+") as json_file:
        com_port = __det_com_port_with_json_file(
            json_cfg_path=json_cfg_path,
            json_file=json_file,
            reconfig_com_port=reconfig_com_port,
        )
    return com_port


def __det_com_port_with_json_file(
    json_cfg_path: str, json_file: TextIO, reconfig_com_port: bool
) -> str:
    try_hint = False
    json_obj = json.load(json_file)
    com_port = ""
    if not reconfig_com_port:
        try_hint, com_port = __try_com_port_load(json_obj=json_obj)
    if try_hint:
        reconfig_com_port, com_port = __try_hint_handling(
            json_cfg_path=json_cfg_path,
            reconfig_com_port=reconfig_com_port,
            json_obj=json_obj,
        )

    if reconfig_com_port:
        com_port = prompt_com_port()
        save_to_json_with_prompt(
            key=JsonKeyNames.SERIAL_PORT.value,
            value=com_port,
            name="serial port",
            json_cfg_path=json_cfg_path,
            json_obj=json_obj,
        )
    json_file.seek(0)
    json.dump(json_obj, json_file, indent=4)
    return com_port


def __try_com_port_load(json_obj) -> Tuple[bool, str]:
    try_hint = False
    com_port = ""
    try:
        com_port = json_obj[JsonKeyNames.SERIAL_PORT.value]
        _LOGGER.info(f"Loaded serial port {com_port} from JSON configuration file")
    except KeyError:
        try_hint = True
    return try_hint, com_port


def __try_hint_handling(json_cfg_path: str, reconfig_com_port: bool, json_obj) -> Tuple[bool, str]:
    reconfig_hint = False
    try:
        hint = json_obj[JsonKeyNames.SERIAL_HINT.value]
    except KeyError:
        reconfig_hint, hint = __prompt_hint_handling(json_obj=json_obj)

    com_port_found, com_port = find_com_port_from_hint(hint=hint)
    if com_port_found:
        _LOGGER.info(f"Found {com_port} based on hint {hint}")
        if reconfig_hint:
            if save_to_json_with_prompt(
                key=JsonKeyNames.SERIAL_PORT.value,
                value=com_port,
                name="serial port",
                json_cfg_path=json_cfg_path,
                json_obj=json_obj,
            ):
                reconfig_com_port = False
    else:
        _LOGGER.info("No serial port found based on hint..")
        reconfig_com_port = True
    return reconfig_com_port, com_port


def __prompt_hint_handling(json_obj) -> Tuple[bool, str]:
    reconfig_hint = False
    hint = ""
    ports = serial.tools.list_ports.comports()
    prompt_hint = input(
        "No hint found in config JSON. Do you want to print the list of devices "
        "and then specify a hint based on it? ([Y]/n): "
    )
    if prompt_hint.lower() in ["y", "yes", "1", ""]:
        while True:
            _LOGGER.info("Found serial devices:")
            for port, desc, hwid in sorted(ports):
                print("{}: {} [{}]".format(port, desc, hwid))
            hint = input("Specify hint: ")
            save_to_json = input(
                "Do you want to store the hint to the configuration file (y) or "
                "specify a new one (r)? ([Y]/r): "
            )
            if save_to_json.lower() in ["y", "yes", "1", ""]:
                json_obj[JsonKeyNames.SERIAL_HINT.value] = hint
                reconfig_hint = True
                break
            elif save_to_json in ["r"]:
                continue
    return reconfig_hint, hint


def find_com_port_from_hint(hint: str) -> Tuple[bool, str]:
    """Find a COM port based on a hint string"""
    if hint == "":
        _LOGGER.warning("Invalid hint, is empty..")
        return False, ""
    ports = serial.tools.list_ports.comports()
    for port, desc, hwid in sorted(ports):
        if hint in desc:
            return True, port
    return False, ""


def prompt_com_port() -> str:
    while True:
        com_port = input(
            "Configuring serial port. Please enter serial port"
            ' or "h" to display list of serial ports): '
        )
        if com_port == "h":
            ports = serial.tools.list_ports.comports()
            for port, desc, hwid in sorted(ports):
                print("{}: {} [{}]".format(port, desc, hwid))
        else:
            if not check_port_validity(com_port):
                print("Serial port not in list of available serial ports. Try again?" " ([Y]/n)")
                try_again = input()
                if try_again.lower() in ["y", "yes", ""]:
                    continue
                else:
                    break
            else:
                break
    return com_port


def check_port_validity(com_port_to_check: str) -> bool:
    port_list = []
    ports = serial.tools.list_ports.comports()
    for port, desc, hwid in sorted(ports):
        port_list.append(port)
    if com_port_to_check not in port_list:
        return False
    return True



import logging
import signal

from tmtccmd.core import BackendBase


def keyboard_interrupt_handler(tmtc_backend: BackendBase):
    tmtc_backend.close_com_if()
    logging.getLogger(__name__).info("Closing TMTC client")


class GracefulKiller:
    kill_now = False

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self):
        self.kill_now = True
        print("I was killed")
